@@ -1,13 +1,19 @@
-/**
+/*
  * XML Security Library (http://www.aleksey.com/xmlsec).
  *
- * Base64 encode/decode transform and utility functions.
  *
  * This is free software; see Copyright file in the source
  * distribution for preciese wording.
  *
- * Copyright (C) 2002-2003 Aleksey Sanin <aleksey@aleksey.com>
+ * Copyright (C) 2002-2022 Aleksey Sanin <aleksey@aleksey.com>. All Rights Reserved.
  */
+/**
+ * SECTION:base64
+ * @Short_description: Base64 encoding/decoding functions and base64 transform implementation.
+ * @Stability: Stable
+ *
+ */
+
 #include "globals.h"
 
 #include <stdlib.h>
@@ -22,6 +28,8 @@
 #include <xmlsec/base64.h>
 #include <xmlsec/errors.h>
 
+#include "cast_helpers.h"
+
 /*
  * the table to map numbers to base64
  */
@@ -45,9 +53,9 @@ static const xmlSecByte base64[] =
 #define xmlSecBase64Encode3(b, c)       ((((b) << 2) & 0x3c) + (((c) >> 6) & 0x03))
 #define xmlSecBase64Encode4(c)          ((c) & 0x3F)
 
-#define xmlSecBase64Decode1(a, b)       (((a) << 2) | (((b) & 0x3F) >> 4))
-#define xmlSecBase64Decode2(b, c)       (((b) << 4) | (((c) & 0x3F) >> 2))
-#define xmlSecBase64Decode3(c, d)       (((c) << 6) | ((d) & 0x3F))
+#define xmlSecBase64Decode1(a, b)       ((xmlSecByte)(((a) << 2) | (((b) & 0x3F) >> 4)))
+#define xmlSecBase64Decode2(b, c)       ((xmlSecByte)(((b) << 4) | (((c) & 0x3F) >> 2)))
+#define xmlSecBase64Decode3(c, d)       ((xmlSecByte)(((c) << 6) | ((d) & 0x3F)))
 
 #define xmlSecIsBase64Char(ch)          ((((ch) >= 'A') && ((ch) <= 'Z')) || \
                                          (((ch) >= 'a') && ((ch) <= 'z')) || \
@@ -73,10 +81,10 @@ typedef enum {
 
 struct _xmlSecBase64Ctx {
     int                 encode;
-    int                 inByte;
-    int                 inPos;
-    xmlSecSize          linePos;
     xmlSecSize          columns;
+    int                 inByte;
+    xmlSecSize          inPos;
+    xmlSecSize          linePos;
     int                 finished;
 };
 
@@ -156,22 +164,13 @@ xmlSecBase64CtxCreate(int encode, int columns) {
      */
     ctx = (xmlSecBase64CtxPtr) xmlMalloc(sizeof(xmlSecBase64Ctx));
     if (ctx == NULL) {
-        xmlSecError(XMLSEC_ERRORS_HERE,
-                    NULL,
-                    NULL,
-                    XMLSEC_ERRORS_R_MALLOC_FAILED,
-                    "sizeof(xmlSecBase64Ctx)=%d",
-                    sizeof(xmlSecBase64Ctx));
+        xmlSecMallocError(sizeof(xmlSecBase64Ctx), NULL);
         return(NULL);
     }
 
     ret = xmlSecBase64CtxInitialize(ctx, encode, columns);
     if(ret < 0) {
-        xmlSecError(XMLSEC_ERRORS_HERE,
-                    NULL,
-                    "xmlSecBase64CtxInitialize",
-                    XMLSEC_ERRORS_R_XMLSEC_FAILED,
-                    XMLSEC_ERRORS_NO_MESSAGE);
+        xmlSecInternalError("xmlSecBase64CtxInitialize", NULL);
         xmlSecBase64CtxDestroy(ctx);
         return(NULL);
     }
@@ -208,8 +207,8 @@ xmlSecBase64CtxInitialize(xmlSecBase64CtxPtr ctx, int encode, int columns) {
 
     memset(ctx, 0, sizeof(xmlSecBase64Ctx));
 
-    ctx->encode     = encode;
-    ctx->columns    = columns;
+    ctx->encode  = encode;
+    XMLSEC_SAFE_CAST_INT_TO_SIZE(columns, ctx->columns, return(-1), NULL);
     return(0);
 }
 
@@ -227,104 +226,87 @@ xmlSecBase64CtxFinalize(xmlSecBase64CtxPtr ctx) {
 }
 
 /**
- * xmlSecBase64CtxUpdate:
+ * xmlSecBase64CtxUpdate_ex:
  * @ctx:                the pointer to #xmlSecBase64Ctx structure
  * @in:                 the input buffer
  * @inSize:             the input buffer size
  * @out:                the output buffer
  * @outSize:            the output buffer size
+ * @outWritten:         the pointer to store the number of bytes written into the output
  *
  * Encodes or decodes the next piece of data from input buffer.
  *
- * Returns: the number of bytes written to output buffer or
- * -1 if an error occurs.
+ * Returns: 0 on success and a negative value otherwise.
  */
 int
-xmlSecBase64CtxUpdate(xmlSecBase64CtxPtr ctx,
-                     const xmlSecByte *in, xmlSecSize inSize,
-                     xmlSecByte *out, xmlSecSize outSize) {
-    xmlSecSize inResSize = 0, outResSize = 0;
+xmlSecBase64CtxUpdate_ex(xmlSecBase64CtxPtr ctx, const xmlSecByte *in, xmlSecSize inSize,
+    xmlSecByte *out, xmlSecSize outSize, xmlSecSize* outWritten) {
+    xmlSecSize inRead = 0;
     int ret;
 
     xmlSecAssert2(ctx != NULL, -1);
     xmlSecAssert2(in != NULL, -1);
     xmlSecAssert2(out != NULL, -1);
+    xmlSecAssert2(outWritten != NULL, -1);
 
     if(ctx->encode != 0) {
-        ret = xmlSecBase64CtxEncode(ctx, in, inSize, &inResSize,
-                                    out, outSize, &outResSize);
-        if((ret < 0) || (inResSize != inSize)) {
-            xmlSecError(XMLSEC_ERRORS_HERE,
-                        NULL,
-                        "xmlSecBase64CtxEncode",
-                        XMLSEC_ERRORS_R_XMLSEC_FAILED,
-                        XMLSEC_ERRORS_NO_MESSAGE);
+        ret = xmlSecBase64CtxEncode(ctx, in, inSize, &inRead, out, outSize, outWritten);
+        if((ret < 0) || (inRead != inSize)) {
+            xmlSecInternalError("xmlSecBase64CtxEncode", NULL);
             return(-1);
         }
     } else {
-        ret = xmlSecBase64CtxDecode(ctx, in, inSize, &inResSize,
-                                    out, outSize, &outResSize);
-        if((ret < 0) || (inResSize != inSize)) {
-            xmlSecError(XMLSEC_ERRORS_HERE,
-                        NULL,
-                        "xmlSecBase64CtxDecode",
-                        XMLSEC_ERRORS_R_XMLSEC_FAILED,
-                        XMLSEC_ERRORS_NO_MESSAGE);
+        ret = xmlSecBase64CtxDecode(ctx, in, inSize, &inRead, out, outSize, outWritten);
+        if((ret < 0) || (inRead != inSize)) {
+            xmlSecInternalError("xmlSecBase64CtxDecode", NULL);
             return(-1);
         }
     }
 
-    return(outResSize);
+    return(0);
 }
 
 /**
- * xmlSecBase64CtxFinal:
+ * xmlSecBase64CtxFinal_ex:
  * @ctx:                the pointer to #xmlSecBase64Ctx structure
  * @out:                the output buffer
  * @outSize:            the output buffer size
+ * @outWritten:         the pointer to store the number of bytes written into the output
  *
  * Encodes or decodes the last piece of data stored in the context
  * and finalizes the result.
  *
- * Returns: the number of bytes written to output buffer or
- * -1 if an error occurs.
+ * Returns: 0 on success and a negative value otherwise.
  */
 int
-xmlSecBase64CtxFinal(xmlSecBase64CtxPtr ctx,
-                    xmlSecByte *out, xmlSecSize outSize) {
-    xmlSecSize outResSize = 0;
-    int ret;
-
+xmlSecBase64CtxFinal_ex(xmlSecBase64CtxPtr ctx, xmlSecByte *out, xmlSecSize outSize, xmlSecSize* outWritten) {
     xmlSecAssert2(ctx != NULL, -1);
     xmlSecAssert2(out != NULL, -1);
     xmlSecAssert2(outSize > 0, -1);
+    xmlSecAssert2(outWritten != NULL, -1);
 
     if(ctx->encode != 0) {
-        ret = xmlSecBase64CtxEncodeFinal(ctx, out, outSize, &outResSize);
+        int ret;
+
+        ret = xmlSecBase64CtxEncodeFinal(ctx, out, outSize, outWritten);
         if(ret < 0) {
-            xmlSecError(XMLSEC_ERRORS_HERE,
-                        NULL,
-                        "xmlSecBase64CtxEncodeFinal",
-                        XMLSEC_ERRORS_R_XMLSEC_FAILED,
-                        "outSize=%d", outSize);
+            xmlSecInternalError2("xmlSecBase64CtxEncodeFinal", NULL,
+                "outSize=" XMLSEC_SIZE_FMT, outSize);
             return(-1);
         }
     } else {
         if(!xmlSecBase64CtxDecodeIsFinished(ctx)) {
-            xmlSecError(XMLSEC_ERRORS_HERE,
-                        NULL,
-                        "xmlSecBase64CtxIsFinished",
-                        XMLSEC_ERRORS_R_XMLSEC_FAILED,
-                        XMLSEC_ERRORS_NO_MESSAGE);
+            xmlSecInternalError("xmlSecBase64CtxDecodeIsFinished", NULL);
             return(-1);
         }
+        (*outWritten) = 0;
     }
 
-    /* add \0 */
-    if((outResSize + 1) < outSize) {
-        out[outResSize] = '\0';
+    /* add \0 just in case (if we can) */
+    if(((*outWritten) + 1) < outSize) {
+        out[(*outWritten)] = '\0';
     }
-    return(outResSize);
+    return(0);
 }
 
 static xmlSecBase64Status
@@ -363,11 +345,7 @@ xmlSecBase64CtxEncodeByte(xmlSecBase64CtxPtr ctx, xmlSecByte inByte, xmlSecByte*
         return(xmlSecBase64StatusConsumeAndNext);
     }
 
-    xmlSecError(XMLSEC_ERRORS_HERE,
-                NULL,
-                NULL,
-                XMLSEC_ERRORS_R_INVALID_DATA,
-                "ctx->inPos=%d", ctx->inPos);
+    xmlSecInvalidSizeDataError("ctx->inPos", ctx->inPos, "0,1,2,3", NULL);
     return(xmlSecBase64StatusFailed);
 }
 
@@ -397,11 +375,7 @@ xmlSecBase64CtxEncodeByteFinal(xmlSecBase64CtxPtr ctx, xmlSecByte* outByte) {
         return(xmlSecBase64StatusConsumeAndRepeat);
     }
 
-    xmlSecError(XMLSEC_ERRORS_HERE,
-                NULL,
-                NULL,
-                XMLSEC_ERRORS_R_INVALID_DATA,
-                "ctx->inPos=%d", ctx->inPos);
+    xmlSecInvalidSizeDataError("ctx->inPos", ctx->inPos, "0,1,2,3", NULL);
     return(xmlSecBase64StatusFailed);
 }
 
@@ -414,49 +388,34 @@ xmlSecBase64CtxDecodeByte(xmlSecBase64CtxPtr ctx, xmlSecByte inByte, xmlSecByte*
         return(xmlSecBase64StatusDone);
     } if(inByte == '=') {
         ctx->finished = 1;
-        if(ctx->inPos < 2) {
-            xmlSecError(XMLSEC_ERRORS_HERE,
-                        NULL,
-                        NULL,
-                        XMLSEC_ERRORS_R_INVALID_DATA,
-                        "ctx->inPos=%d", ctx->inPos);
-            return(xmlSecBase64StatusFailed);
-        } else if(ctx->inPos == 2) {
+        if(ctx->inPos == 2) {
             ++ctx->inPos;
             return(xmlSecBase64StatusNext);
         } else if(ctx->inPos == 3) {
             ctx->inPos = 0;
             return(xmlSecBase64StatusNext);
         } else {
-            xmlSecError(XMLSEC_ERRORS_HERE,
-                        NULL,
-                        NULL,
-                        XMLSEC_ERRORS_R_INVALID_DATA,
-                        "ctx->inPos=%d", ctx->inPos);
+            xmlSecInvalidSizeDataError("ctx->inPos", ctx->inPos, "2,3", NULL);
             return(xmlSecBase64StatusFailed);
         }
     } else if(xmlSecIsBase64Space(inByte)) {
         return(xmlSecBase64StatusNext);
     } else if(!xmlSecIsBase64Char(inByte) || (ctx->finished != 0)) {
-        xmlSecError(XMLSEC_ERRORS_HERE,
-                    NULL,
-                    NULL,
-                    XMLSEC_ERRORS_R_INVALID_DATA,
-                    "inByte=0x%02x", inByte);
+        xmlSecInvalidIntegerDataError("inByte", inByte, "base64 character", NULL);
         return(xmlSecBase64StatusFailed);
     }
 
     /* convert from character to position in base64 array */
     if((inByte >= 'A') && (inByte <= 'Z')) {
-        inByte = (inByte - 'A');
+        inByte = (xmlSecByte)(inByte - 'A');
     } else if((inByte >= 'a') && (inByte <= 'z')) {
-        inByte = 26 + (inByte - 'a');
+        inByte = (xmlSecByte)(26 + (inByte - 'a'));
     } else if((inByte >= '0') && (inByte <= '9')) {
-        inByte = 52 + (inByte - '0');
+        inByte = (xmlSecByte)(52 + (inByte - '0'));
     } else if(inByte == '+') {
-        inByte = 62;
+        inByte = (xmlSecByte)62;
     } else if(inByte == '/') {
-        inByte = 63;
+        inByte = (xmlSecByte)63;
     }
 
     if(ctx->inPos == 0) {
@@ -480,11 +439,7 @@ xmlSecBase64CtxDecodeByte(xmlSecBase64CtxPtr ctx, xmlSecByte inByte, xmlSecByte*
         return(xmlSecBase64StatusConsumeAndNext);
     }
 
-    xmlSecError(XMLSEC_ERRORS_HERE,
-                NULL,
-                NULL,
-                XMLSEC_ERRORS_R_INVALID_DATA,
-                "ctx->inPos=%d", ctx->inPos);
+    xmlSecInvalidSizeDataError("ctx->inPos", ctx->inPos, "0,1,2,3", NULL);
     return(xmlSecBase64StatusFailed);
 }
 
@@ -516,11 +471,8 @@ xmlSecBase64CtxEncode(xmlSecBase64CtxPtr ctx,
             case xmlSecBase64StatusNext:
             case xmlSecBase64StatusDone:
             case xmlSecBase64StatusFailed:
-                xmlSecError(XMLSEC_ERRORS_HERE,
-                            NULL,
-                            "xmlSecBase64CtxEncodeByte",
-                            XMLSEC_ERRORS_R_XMLSEC_FAILED,
-                            "status=%d", status);
+                xmlSecInternalError2("xmlSecBase64CtxEncodeByte", NULL,
+                    "status=" XMLSEC_ENUM_FMT, XMLSEC_ENUM_CAST(status));
                 return(-1);
         }
     }
@@ -532,8 +484,9 @@ xmlSecBase64CtxEncode(xmlSecBase64CtxPtr ctx,
 }
 
 static int
-xmlSecBase64CtxEncodeFinal(xmlSecBase64CtxPtr ctx,
-                     xmlSecByte* outBuf, xmlSecSize outBufSize, xmlSecSize* outBufResSize) {
+xmlSecBase64CtxEncodeFinal(xmlSecBase64CtxPtr ctx, xmlSecByte* outBuf, xmlSecSize outBufSize,
+    xmlSecSize* outBufResSize) {
+
     xmlSecBase64Status status = xmlSecBase64StatusNext;
     xmlSecSize outPos;
 
@@ -553,21 +506,14 @@ xmlSecBase64CtxEncodeFinal(xmlSecBase64CtxPtr ctx,
                 break;
             case xmlSecBase64StatusNext:
             case xmlSecBase64StatusFailed:
-                xmlSecError(XMLSEC_ERRORS_HERE,
-                            NULL,
-                            "xmlSecBase64CtxEncodeByteFinal",
-                            XMLSEC_ERRORS_R_XMLSEC_FAILED,
-                            "status=%d", status);
+                xmlSecInternalError2("xmlSecBase64CtxEncodeByteFinal", NULL,
+                    "status=" XMLSEC_ENUM_FMT, XMLSEC_ENUM_CAST(status));
                 return(-1);
         }
     }
 
     if(status != xmlSecBase64StatusDone) {
-        xmlSecError(XMLSEC_ERRORS_HERE,
-                    NULL,
-                    NULL,
-                    XMLSEC_ERRORS_R_INVALID_SIZE,
-                    "outBufSize=%d", outBufSize);
+        xmlSecInvalidSizeOtherError("invalid base64 buffer size", NULL);
         return(-1);
     }
     if(outPos < outBufSize) {
@@ -609,11 +555,8 @@ xmlSecBase64CtxDecode(xmlSecBase64CtxPtr ctx,
             case xmlSecBase64StatusDone:
                 break;
             case xmlSecBase64StatusFailed:
-                xmlSecError(XMLSEC_ERRORS_HERE,
-                            NULL,
-                            "xmlSecBase64CtxDecodeByte",
-                            XMLSEC_ERRORS_R_XMLSEC_FAILED,
-                            "status=%d", status);
+                xmlSecInternalError2("xmlSecBase64CtxDecodeByte", NULL,
+                    "status=" XMLSEC_ENUM_FMT, XMLSEC_ENUM_CAST(status));
                 return(-1);
         }
     }
@@ -636,10 +579,23 @@ xmlSecBase64CtxDecodeIsFinished(xmlSecBase64CtxPtr ctx) {
     return((ctx->inPos == 0) ? 1 : 0);
 }
 
+static xmlSecSize
+xmlSecBase64GetEncodeSize(xmlSecBase64CtxPtr ctx, xmlSecSize inLen) {
+    xmlSecSize size;
+
+    xmlSecAssert2(ctx != NULL, 0);
+
+    size = (4 * inLen) / 3 + 4;
+    if(ctx->columns > 0) {
+        size += (size / ctx->columns) + 4;
+    }
+    return(size + 1);
+}
+
 /**
  * xmlSecBase64Encode:
- * @buf:                the input buffer.
- * @len:                the input buffer size.
+ * @in:                 the input buffer.
+ * @inSize:             the input buffer size.
  * @columns:            the output max line length (if 0 then no line breaks
  *                      would be inserted)
  *
@@ -651,145 +607,236 @@ xmlSecBase64CtxDecodeIsFinished(xmlSecBase64CtxPtr ctx) {
  * or NULL if an error occurs.
  */
 xmlChar*
-xmlSecBase64Encode(const xmlSecByte *buf, xmlSecSize len, int columns) {
+xmlSecBase64Encode(const xmlSecByte *in, xmlSecSize inSize, int columns) {
     xmlSecBase64Ctx ctx;
-    xmlChar *ptr;
-    xmlSecSize size;
-    int size_update, size_final;
+    int ctx_initialized = 0;
+    xmlSecByte* ptr = NULL;
+    xmlChar* res = NULL;
+    xmlSecSize outSize, outUpdatedSize, outFinalSize;
     int ret;
 
-    xmlSecAssert2(buf != NULL, NULL);
+    xmlSecAssert2(in != NULL, NULL);
 
     ret = xmlSecBase64CtxInitialize(&ctx, 1, columns);
     if(ret < 0) {
-        xmlSecError(XMLSEC_ERRORS_HERE,
-                    NULL,
-                    "xmlSecBase64CtxInitialize",
-                    XMLSEC_ERRORS_R_XMLSEC_FAILED,
-                    XMLSEC_ERRORS_NO_MESSAGE);
-        return(NULL);
+        xmlSecInternalError("xmlSecBase64CtxInitialize", NULL);
+        goto done;
     }
+    ctx_initialized = 1;
 
     /* create result buffer */
-    size = (4 * len) / 3 + 4;
-    if(columns > 0) {
-        size += (size / columns) + 4;
+    outSize = xmlSecBase64GetEncodeSize(&ctx, inSize);
+    if(outSize == 0) {
+        xmlSecInternalError("xmlSecBase64GetEncodeSize", NULL);
+        goto done;
     }
-    ptr = (xmlChar*) xmlMalloc(size);
+    ptr = (xmlSecByte*)xmlMalloc(outSize);
     if(ptr == NULL) {
-        xmlSecError(XMLSEC_ERRORS_HERE,
-                    NULL,
-                    NULL,
-                    XMLSEC_ERRORS_R_MALLOC_FAILED,
-                    "size=%d", size);
-        xmlSecBase64CtxFinalize(&ctx);
-        return(NULL);
+        xmlSecMallocError(outSize, NULL);
+        goto done;
     }
 
-    ret = xmlSecBase64CtxUpdate(&ctx, buf, len, (xmlSecByte*)ptr, size);
-    if(ret < 0) {
-        xmlSecError(XMLSEC_ERRORS_HERE,
-                    NULL,
-                    "xmlSecBase64CtxUpdate",
-                    XMLSEC_ERRORS_R_XMLSEC_FAILED,
-                    "len=%d", len);
+    ret = xmlSecBase64CtxUpdate_ex(&ctx, in, inSize, ptr, outSize, &outUpdatedSize);
+    if (ret < 0) {
+        xmlSecInternalError3("xmlSecBase64CtxUpdate_ex", NULL,
+            "inSize=" XMLSEC_SIZE_FMT "; outSize=" XMLSEC_SIZE_FMT, inSize, outSize);
+        goto done;
+    }
+
+    ret = xmlSecBase64CtxFinal_ex(&ctx, ptr + outUpdatedSize, outSize - outUpdatedSize,
+        &outFinalSize);
+    if (ret < 0) {
+        xmlSecInternalError("xmlSecBase64CtxFinal_ex", NULL);
+        goto done;
+    }
+
+    /* success */
+    ptr[outUpdatedSize + outFinalSize] = '\0';
+    res = BAD_CAST(ptr);
+    ptr = NULL;
+
+done:
+    if(ptr != NULL) {
         xmlFree(ptr);
-        xmlSecBase64CtxFinalize(&ctx);
-        return(NULL);
     }
-    size_update = ret;
-
-    ret = xmlSecBase64CtxFinal(&ctx, ((xmlSecByte*)ptr) + size_update, size - size_update);
-    if(ret < 0) {
-        xmlSecError(XMLSEC_ERRORS_HERE,
-                    NULL,
-                    "xmlSecBase64CtxFinal",
-                    XMLSEC_ERRORS_R_XMLSEC_FAILED,
-                    XMLSEC_ERRORS_NO_MESSAGE);
-        xmlFree(ptr);
+    if(ctx_initialized != 0) {
         xmlSecBase64CtxFinalize(&ctx);
-        return(NULL);
     }
-    size_final = ret;
-    ptr[size_update + size_final] = '\0';
-
-    xmlSecBase64CtxFinalize(&ctx);
-    return(ptr);
+    return(res);
 }
 
 /**
- * xmlSecBase64Decode:
+ * xmlSecBase64Decode_ex:
  * @str:                the input buffer with base64 encoded string
- * @buf:                the output buffer
- * @len:                the output buffer size
+ * @out:                the output buffer
+ * @outSize:            the output buffer size
+ * @outWritten:         the pointer to store the number of bytes written into the output.
  *
  * Decodes input base64 encoded string and puts result into
  * the output buffer.
  *
- * Returns: the number of bytes written to the output buffer or
- * a negative value if an error occurs
+ * Returns: 0 on success and a negative value otherwise.
  */
 int
-xmlSecBase64Decode(const xmlChar* str, xmlSecByte *buf, xmlSecSize len) {
+xmlSecBase64Decode_ex(const xmlChar* str, xmlSecByte* out, xmlSecSize outSize, xmlSecSize* outWritten) {
     xmlSecBase64Ctx ctx;
-    int size_update;
-    int size_final;
+    int ctx_initialized = 0;
+    xmlSecSize outUpdateSize, outFinalSize;
     int ret;
+    int res = -1;
 
     xmlSecAssert2(str != NULL, -1);
-    xmlSecAssert2(buf != NULL, -1);
+    xmlSecAssert2(out != NULL, -1);
+    xmlSecAssert2(outWritten != NULL, -1);
 
     ret = xmlSecBase64CtxInitialize(&ctx, 0, 0);
     if(ret < 0) {
-        xmlSecError(XMLSEC_ERRORS_HERE,
-                    NULL,
-                    "xmlSecBase64CtxInitialize",
-                    XMLSEC_ERRORS_R_XMLSEC_FAILED,
-                    XMLSEC_ERRORS_NO_MESSAGE);
-        return(-1);
+        xmlSecInternalError("xmlSecBase64CtxInitialize", NULL);
+        goto done;
+    }
+    ctx_initialized = 1;
+
+    ret = xmlSecBase64CtxUpdate_ex(&ctx, (const xmlSecByte*)str, xmlSecStrlen(str),
+        out, outSize, &outUpdateSize);
+    if (ret < 0) {
+        xmlSecInternalError("xmlSecBase64CtxUpdate_ex", NULL);
+        goto done;
     }
 
-    ret = xmlSecBase64CtxUpdate(&ctx, (const xmlSecByte*)str, xmlStrlen(str), buf, len);
-    if(ret < 0) {
-        xmlSecError(XMLSEC_ERRORS_HERE,
-                    NULL,
-                    "xmlSecBase64CtxUpdate",
-                    XMLSEC_ERRORS_R_XMLSEC_FAILED,
-                    XMLSEC_ERRORS_NO_MESSAGE);
+    ret = xmlSecBase64CtxFinal_ex(&ctx, out + outUpdateSize, outSize - outUpdateSize,
+        &outFinalSize);
+    if (ret < 0) {
+        xmlSecInternalError("xmlSecBase64CtxFinal_ex", NULL);
+        goto done;
+    }
+
+    /* success */
+    (*outWritten) = (outUpdateSize + outFinalSize);
+    res = 0;
+
+done:
+    if(ctx_initialized != 0) {
         xmlSecBase64CtxFinalize(&ctx);
-        return(-1);
     }
-
-    size_update = ret;
-    ret = xmlSecBase64CtxFinal(&ctx, buf + size_update, len - size_update);
-    if(ret < 0) {
-        xmlSecError(XMLSEC_ERRORS_HERE,
-                    NULL,
-                    "xmlSecBase64CtxFinal",
-                    XMLSEC_ERRORS_R_XMLSEC_FAILED,
-                    XMLSEC_ERRORS_NO_MESSAGE);
-        xmlSecBase64CtxFinalize(&ctx);
-        return(-1);
-    }
-    size_final = ret;
-
-    xmlSecBase64CtxFinalize(&ctx);
-    return(size_update + size_final);
+    return(res);
 }
 
+/**
+ * xmlSecBase64DecodeInPlace:
+ * @str:                the input/output buffer
+ * @outWritten:         the pointer to store the number of bytes written into the output.
+ *
+ * Decodes input base64 encoded string from @str "in-place" (i.e. puts results into @str buffer).
+ *
+ * Returns: 0 on success and a negative value otherwise.
+ */
+int
+xmlSecBase64DecodeInPlace(xmlChar* str, xmlSecSize* outWritten) {
+    xmlSecAssert2(str != NULL, -1);
+    return(xmlSecBase64Decode_ex(str, (xmlSecByte*)str, xmlSecStrlen(str) + 1,outWritten));
+}
+
+/**************************************************************
+ *
+ * Deprecated functions for backward compatibility
+ *
+ **************************************************************/
+
+/**
+ * xmlSecBase64CtxUpdate:
+ * @ctx:                the pointer to #xmlSecBase64Ctx structure
+ * @in:                 the input buffer
+ * @inSize:             the input buffer size
+ * @out:                the output buffer
+ * @outSize:            the output buffer size
+ *
+ * DEPRECATED. Encodes or decodes the next piece of data from input buffer.
+ *
+ * Returns: the number of bytes written to output buffer or
+ * -1 if an error occurs.
+ */
+int
+xmlSecBase64CtxUpdate(xmlSecBase64CtxPtr ctx, const xmlSecByte* in, xmlSecSize inSize,
+    xmlSecByte* out, xmlSecSize outSize) {
+
+    int ret;
+    xmlSecSize outWritten;
+    int res;
+
+    ret = xmlSecBase64CtxUpdate_ex(ctx, in, inSize, out, outSize, &outWritten);
+    if (ret < 0) {
+        return(-1);
+    }
+
+    XMLSEC_SAFE_CAST_SIZE_TO_INT(outWritten, res, return(-1), NULL);
+    return(res);
+}
+
+
+/**
+ * xmlSecBase64CtxFinal:
+ * @ctx:                the pointer to #xmlSecBase64Ctx structure
+ * @out:                the output buffer
+ * @outSize:            the output buffer size
+ *
+ * DEPRECATED. Encodes or decodes the last piece of data stored in the context
+ * and finalizes the result.
+ *
+ * Returns: the number of bytes written to output buffer or
+ * -1 if an error occurs.
+ */
+int
+xmlSecBase64CtxFinal(xmlSecBase64CtxPtr ctx, xmlSecByte* out, xmlSecSize outSize) {
+    int ret;
+    xmlSecSize outWritten;
+    int res;
+
+    ret = xmlSecBase64CtxFinal_ex(ctx, out, outSize, &outWritten);
+    if (ret < 0) {
+        return(-1);
+    }
+
+    XMLSEC_SAFE_CAST_SIZE_TO_INT(outWritten, res, return(-1), NULL);
+    return(res);
+}
+
+ /**
+  * xmlSecBase64Decode:
+  * @str:                the input buffer with base64 encoded string
+  * @out:                the output buffer
+  * @outSize:            the output buffer size
+  *
+  * DEPRECATED. Decodes input base64 encoded string and puts result into
+  * the output buffer.
+  *
+  * Returns: the number of bytes written to the output buffer or
+  * a negative value if an error occurs
+  */
+int
+xmlSecBase64Decode(const xmlChar* str, xmlSecByte* out, xmlSecSize outSize) {
+    int ret;
+    xmlSecSize outWritten;
+    int res;
+
+    ret = xmlSecBase64Decode_ex(str, out, outSize, &outWritten);
+    if (ret < 0) {
+        return(-1);
+    }
+
+    XMLSEC_SAFE_CAST_SIZE_TO_INT(outWritten, res, return(-1), NULL);
+    return(res);
+}
+
+
 /**************************************************************
  *
  * Base64 Transform
  *
- * xmlSecBase64Ctx is located after xmlSecTransform
+ * xmlSecTransform + xmlSecBase64Ctx
  *
  **************************************************************/
-#define xmlSecBase64Size \
-        (sizeof(xmlSecTransform) + sizeof(xmlSecBase64Ctx))
-#define xmlSecBase64GetCtx(transform) \
-    ((xmlSecTransformCheckSize((transform), xmlSecBase64Size)) ? \
-        (xmlSecBase64CtxPtr)(((xmlSecByte*)(transform)) + sizeof(xmlSecTransform)) : \
-        (xmlSecBase64CtxPtr)NULL)
+XMLSEC_TRANSFORM_DECLARE(Base64, xmlSecBase64Ctx)
+#define xmlSecBase64Size XMLSEC_TRANSFORM_SIZE(Base64)
 
 static int              xmlSecBase64Initialize          (xmlSecTransformPtr transform);
 static void             xmlSecBase64Finalize            (xmlSecTransformPtr transform);
@@ -864,20 +911,18 @@ static int
 xmlSecBase64Initialize(xmlSecTransformPtr transform) {
     xmlSecBase64CtxPtr ctx;
     int ret;
+    int columns;
 
     xmlSecAssert2(xmlSecTransformCheckId(transform, xmlSecTransformBase64Id), -1);
 
     ctx = xmlSecBase64GetCtx(transform);
     xmlSecAssert2(ctx != NULL, -1);
 
+    columns = xmlSecBase64GetDefaultLineSize();
     transform->operation = xmlSecTransformOperationDecode;
-    ret = xmlSecBase64CtxInitialize(ctx, 0, xmlSecBase64GetDefaultLineSize());
+    ret = xmlSecBase64CtxInitialize(ctx, 0, columns);
     if(ret < 0) {
-        xmlSecError(XMLSEC_ERRORS_HERE,
-                    xmlSecErrorsSafeString(xmlSecTransformGetName(transform)),
-                    "xmlSecBase64CtxInitialize",
-                    XMLSEC_ERRORS_R_XMLSEC_FAILED,
-                    XMLSEC_ERRORS_NO_MESSAGE);
+        xmlSecInternalError("xmlSecBase64CtxInitialize", xmlSecTransformGetName(transform));
         return(-1);
     }
 
@@ -900,7 +945,7 @@ static int
 xmlSecBase64Execute(xmlSecTransformPtr transform, int last, xmlSecTransformCtxPtr transformCtx) {
     xmlSecBase64CtxPtr ctx;
     xmlSecBufferPtr in, out;
-    xmlSecSize inSize, outSize, outLen;
+    xmlSecSize inSize, outSize, outMaxLen, outLen;
     int ret;
 
     xmlSecAssert2(xmlSecTransformCheckId(transform, xmlSecTransformBase64Id), -1);
@@ -924,93 +969,69 @@ xmlSecBase64Execute(xmlSecTransformPtr transform, int last, xmlSecTransformCtxPt
             outSize = xmlSecBufferGetSize(out);
             if(inSize > 0) {
                 if(ctx->encode != 0) {
-                    outLen = 4 * inSize / 3 + 8;
+                    outMaxLen = 4 * inSize / 3 + 8;
                     if(ctx->columns > 0) {
-                        outLen += inSize / ctx->columns + 4;
+                        outMaxLen += inSize / ctx->columns + 4;
                     }
                 } else {
-                    outLen = 3 * inSize / 4 + 8;
+                    outMaxLen = 3 * inSize / 4 + 8;
                 }
-                ret = xmlSecBufferSetMaxSize(out, outSize + outLen);
+                ret = xmlSecBufferSetMaxSize(out, outSize + outMaxLen);
                 if(ret < 0) {
-                    xmlSecError(XMLSEC_ERRORS_HERE,
-                                xmlSecErrorsSafeString(xmlSecTransformGetName(transform)),
-                                "xmlSecBufferSetMaxSize",
-                                XMLSEC_ERRORS_R_XMLSEC_FAILED,
-                                "size=%d", outSize + outLen);
+                    xmlSecInternalError2("xmlSecBufferSetMaxSize", xmlSecTransformGetName(transform),
+                        "size=" XMLSEC_SIZE_FMT, (outSize + outMaxLen));
                     return(-1);
                 }
 
                 /* encode/decode the next chunk */
-                ret = xmlSecBase64CtxUpdate(ctx, xmlSecBufferGetData(in), inSize,
-                                            xmlSecBufferGetData(out) + outSize,
-                                            outLen);
+                ret = xmlSecBase64CtxUpdate_ex(ctx, xmlSecBufferGetData(in), inSize,
+                    xmlSecBufferGetData(out) + outSize, outMaxLen, &outLen);
                 if(ret < 0) {
-                    xmlSecError(XMLSEC_ERRORS_HERE,
-                                xmlSecErrorsSafeString(xmlSecTransformGetName(transform)),
-                                "xmlSecBase64CtxUpdate",
-                                XMLSEC_ERRORS_R_XMLSEC_FAILED,
-                                XMLSEC_ERRORS_NO_MESSAGE);
+                    xmlSecInternalError("xmlSecBase64CtxUpdate_ex", xmlSecTransformGetName(transform));
                     return(-1);
                 }
-                outLen = ret;
 
                 /* set correct size */
                 ret = xmlSecBufferSetSize(out, outSize + outLen);
                 if(ret < 0) {
-                    xmlSecError(XMLSEC_ERRORS_HERE,
-                                xmlSecErrorsSafeString(xmlSecTransformGetName(transform)),
-                                "xmlSecBufferSetSize",
-                                XMLSEC_ERRORS_R_XMLSEC_FAILED,
-                                "size=%d", outSize + outLen);
+                    xmlSecInternalError2("xmlSecBufferSetSize", xmlSecTransformGetName(transform),
+                        "size=" XMLSEC_SIZE_FMT, (outSize + outLen));
                     return(-1);
                 }
 
                 /* remove chunk from input */
                 ret = xmlSecBufferRemoveHead(in, inSize);
                 if(ret < 0) {
-                    xmlSecError(XMLSEC_ERRORS_HERE,
-                                xmlSecErrorsSafeString(xmlSecTransformGetName(transform)),
-                                "xmlSecBufferRemoveHead",
-                                XMLSEC_ERRORS_R_XMLSEC_FAILED,
-                                "size=%d", inSize);
+                    xmlSecInternalError2("xmlSecBufferRemoveHead", xmlSecTransformGetName(transform),
+                        "size=" XMLSEC_SIZE_FMT, inSize);
                     return(-1);
                 }
             }
 
             if(last) {
                 outSize = xmlSecBufferGetSize(out);
+                outMaxLen = 16; /* last block */
 
-                ret = xmlSecBufferSetMaxSize(out, outSize + 16);
+                ret = xmlSecBufferSetMaxSize(out, outSize + outMaxLen);
                 if(ret < 0) {
-                    xmlSecError(XMLSEC_ERRORS_HERE,
-                                xmlSecErrorsSafeString(xmlSecTransformGetName(transform)),
-                                "xmlSecBufferSetMaxSize",
-                                XMLSEC_ERRORS_R_XMLSEC_FAILED,
-                                "size=%d", outSize + 16);
+                    xmlSecInternalError2("xmlSecBufferSetMaxSize", xmlSecTransformGetName(transform),
+                        "size=" XMLSEC_SIZE_FMT, (outSize + outMaxLen));
                     return(-1);
                 }
 
                 /* add from ctx buffer */
-                ret = xmlSecBase64CtxFinal(ctx, xmlSecBufferGetData(out) + outSize, 16);
-                if(ret < 0) {
-                    xmlSecError(XMLSEC_ERRORS_HERE,
-                                xmlSecErrorsSafeString(xmlSecTransformGetName(transform)),
-                                "xmlSecBase64CtxFinal",
-                                XMLSEC_ERRORS_R_XMLSEC_FAILED,
-                                XMLSEC_ERRORS_NO_MESSAGE);
+                ret = xmlSecBase64CtxFinal_ex(ctx, xmlSecBufferGetData(out) + outSize,
+                    outMaxLen, &outLen);
+                if (ret < 0) {
+                    xmlSecInternalError("xmlSecBase64CtxFinal_ex", xmlSecTransformGetName(transform));
                     return(-1);
                 }
-                outLen = ret;
 
                 /* set correct size */
                 ret = xmlSecBufferSetSize(out, outSize + outLen);
                 if(ret < 0) {
-                    xmlSecError(XMLSEC_ERRORS_HERE,
-                                xmlSecErrorsSafeString(xmlSecTransformGetName(transform)),
-                                "xmlSecBufferSetSize",
-                                XMLSEC_ERRORS_R_XMLSEC_FAILED,
-                                "size=%d", outSize + outLen);
+                    xmlSecInternalError2("xmlSecBufferSetSize", xmlSecTransformGetName(transform),
+                        "size=" XMLSEC_SIZE_FMT, (outSize + outLen));
                     return(-1);
                 }
                 transform->status = xmlSecTransformStatusFinished;
@@ -1021,11 +1042,7 @@ xmlSecBase64Execute(xmlSecTransformPtr transform, int last, xmlSecTransformCtxPt
             xmlSecAssert2(xmlSecBufferGetSize(in) == 0, -1);
             break;
         default:
-            xmlSecError(XMLSEC_ERRORS_HERE,
-                        xmlSecErrorsSafeString(xmlSecTransformGetName(transform)),
-                        NULL,
-                        XMLSEC_ERRORS_R_INVALID_STATUS,
-                        "status=%d", transform->status);
+            xmlSecInvalidTransfromStatusError(transform);
             return(-1);
     }
     return(0);
