@@ -5,7 +5,7 @@
  * This is free software; see Copyright file in the source
  * distribution for preciese wording.
  *
- * Copyright (C) 2010-2016 Aleksey Sanin <aleksey@aleksey.com>. All Rights Reserved.
+ * Copyright (C) 2002-2022 Aleksey Sanin <aleksey@aleksey.com>. All Rights Reserved.
  */
 /**
  * SECTION:x509vfy
@@ -24,20 +24,19 @@
 #include <ctype.h>
 #include <errno.h>
 
-#include <libxml/tree.h>
-
 #include <xmlsec/xmlsec.h>
-#include <xmlsec/xmltree.h>
 #include <xmlsec/keys.h>
 #include <xmlsec/keyinfo.h>
 #include <xmlsec/keysmngr.h>
 #include <xmlsec/base64.h>
 #include <xmlsec/errors.h>
+#include <xmlsec/private.h>
 
 #include <xmlsec/gnutls/crypto.h>
 #include <xmlsec/gnutls/x509.h>
 
 #include "x509utils.h"
+#include "../cast_helpers.h"
 
 /**************************************************************************
  *
@@ -55,14 +54,9 @@ struct _xmlSecGnuTLSX509StoreCtx {
  *
  * xmlSecGnuTLSKeyDataStoreX509Id:
  *
- * xmlSecGnuTLSX509StoreCtx is located after xmlSecTransform
- *
  ***************************************************************************/
-#define xmlSecGnuTLSX509StoreGetCtx(store) \
-    ((xmlSecGnuTLSX509StoreCtxPtr)(((xmlSecByte*)(store)) + \
-                                    sizeof(xmlSecKeyDataStoreKlass)))
-#define xmlSecGnuTLSX509StoreSize      \
-    (sizeof(xmlSecKeyDataStoreKlass) + sizeof(xmlSecGnuTLSX509StoreCtx))
+XMLSEC_KEY_DATA_STORE_DECLARE(GnuTLSX509Store, xmlSecGnuTLSX509StoreCtx)
+#define xmlSecGnuTLSX509StoreSize XMLSEC_KEY_DATA_STORE_SIZE(GnuTLSX509Store)
 
 static int              xmlSecGnuTLSX509StoreInitialize                 (xmlSecKeyDataStorePtr store);
 static void             xmlSecGnuTLSX509StoreFinalize                   (xmlSecKeyDataStorePtr store);
@@ -87,13 +81,17 @@ static gnutls_x509_crt_t xmlSecGnuTLSX509FindCert                       (xmlSecP
                                                                          const xmlChar *subjectName,
                                                                          const xmlChar *issuerName,
                                                                          const xmlChar *issuerSerial,
-                                                                         const xmlChar *ski);
+                                                                         const xmlSecByte * ski,
+                                                                         xmlSecSize skiSize);
 static gnutls_x509_crt_t xmlSecGnuTLSX509FindSignedCert                 (xmlSecPtrListPtr certs,
                                                                          gnutls_x509_crt_t cert);
 static gnutls_x509_crt_t xmlSecGnuTLSX509FindSignerCert                 (xmlSecPtrListPtr certs,
                                                                          gnutls_x509_crt_t cert);
 
 
+static int               xmlSecGnuTLSX509CertCompareSKI                 (gnutls_x509_crt_t cert,
+                                                                         const xmlSecByte * ski,
+                                                                         xmlSecSize skiSize);
 /**
  * xmlSecGnuTLSX509StoreGetKlass:
  *
@@ -121,33 +119,86 @@ xmlSecGnuTLSX509StoreGetKlass(void) {
  * or an error occurs.
  */
 gnutls_x509_crt_t
-xmlSecGnuTLSX509StoreFindCert(xmlSecKeyDataStorePtr store,
-                              const xmlChar *subjectName,
-                              const xmlChar *issuerName,
-                              const xmlChar *issuerSerial,
-                              const xmlChar *ski,
-                              const xmlSecKeyInfoCtx* keyInfoCtx) {
+xmlSecGnuTLSX509StoreFindCert(const xmlSecKeyDataStorePtr store, const xmlChar *subjectName,
+                              const xmlChar *issuerName, const xmlChar *issuerSerial,
+                              const xmlChar *ski, const xmlSecKeyInfoCtx* keyInfoCtx ) {
+    if(ski != NULL) {
+        gnutls_x509_crt_t res;
+        xmlChar* skiDup;
+        xmlSecSize skiDecodedSize = 0;
+        int ret;
+
+        skiDup = xmlStrdup(ski);
+        if(skiDup == NULL) {
+            xmlSecStrdupError(ski, NULL);
+            return(NULL);
+        }
+
+        /* our usual trick with base64 decode */
+        ret = xmlSecBase64DecodeInPlace(skiDup, &skiDecodedSize);
+        if(ret < 0) {
+            xmlSecInternalError2("xmlSecBase64DecodeInPlace", NULL,
+                "ski=%s", xmlSecErrorsSafeString(skiDup));
+            xmlFree(skiDup);
+            return(NULL);
+        }
+
+        res = xmlSecGnuTLSX509StoreFindCert_ex(store, subjectName, issuerName, issuerSerial,
+            (xmlSecByte*)skiDup, skiDecodedSize, keyInfoCtx);
+        xmlFree(skiDup);
+        return(res);
+    } else {
+        return(xmlSecGnuTLSX509StoreFindCert_ex(store, subjectName, issuerName, issuerSerial,
+            NULL, 0, keyInfoCtx));
+
+    }
+}
+
+/**
+ * xmlSecGnuTLSX509StoreFindCert_ex:
+ * @store:              the pointer to X509 key data store klass.
+ * @subjectName:        the desired certificate name.
+ * @issuerName:         the desired certificate issuer name.
+ * @issuerSerial:       the desired certificate issuer serial number.
+ * @ski:                the desired certificate SKI.
+ * @skiSize:            the desired certificate SKI size.
+ * @keyInfoCtx:         the pointer to <dsig:KeyInfo/> element processing context.
+ *
+ * Searches @store for a certificate that matches given criteria.
+ *
+ * Returns: pointer to found certificate or NULL if certificate is not found
+ * or an error occurs.
+ */
+gnutls_x509_crt_t
+xmlSecGnuTLSX509StoreFindCert_ex(const xmlSecKeyDataStorePtr store, const xmlChar *subjectName,
+                              const xmlChar *issuerName, const xmlChar *issuerSerial,
+                              const xmlSecByte * ski, xmlSecSize skiSize,
+                              const xmlSecKeyInfoCtx* keyInfoCtx ATTRIBUTE_UNUSED) {
     xmlSecGnuTLSX509StoreCtxPtr ctx;
     gnutls_x509_crt_t res = NULL;
 
     xmlSecAssert2(xmlSecKeyDataStoreCheckId(store, xmlSecGnuTLSX509StoreId), NULL);
-    xmlSecAssert2(keyInfoCtx != NULL, NULL);
+    UNREFERENCED_PARAMETER(keyInfoCtx);
 
     ctx = xmlSecGnuTLSX509StoreGetCtx(store);
     xmlSecAssert2(ctx != NULL, NULL);
 
     if(res == NULL) {
-        res = xmlSecGnuTLSX509FindCert(&(ctx->certsTrusted), subjectName, issuerName, issuerSerial, ski);
+        res = xmlSecGnuTLSX509FindCert(&(ctx->certsTrusted), subjectName,
+            issuerName, issuerSerial,
+            ski, skiSize);
     }
     if(res == NULL) {
-        res = xmlSecGnuTLSX509FindCert(&(ctx->certsUntrusted), subjectName, issuerName, issuerSerial, ski);
+        res = xmlSecGnuTLSX509FindCert(&(ctx->certsUntrusted), subjectName,
+            issuerName, issuerSerial,
+            ski, skiSize);
     }
     return(res);
 }
 
 static int
 xmlSecGnuTLSX509CheckTime(const gnutls_x509_crt_t * cert_list,
-                          xmlSecSize cert_list_length,
+                          xmlSecSize cert_list_size,
                           time_t ts)
 {
     time_t notValidBefore, notValidAfter;
@@ -155,7 +206,7 @@ xmlSecGnuTLSX509CheckTime(const gnutls_x509_crt_t * cert_list,
 
     xmlSecAssert2(cert_list != NULL, -1);
 
-    for(ii = 0; ii < cert_list_length; ++ii) {
+    for(ii = 0; ii < cert_list_size; ++ii) {
         const gnutls_x509_crt_t cert = cert_list[ii];
         if(cert == NULL) {
             continue;
@@ -165,17 +216,17 @@ xmlSecGnuTLSX509CheckTime(const gnutls_x509_crt_t * cert_list,
         notValidBefore = gnutls_x509_crt_get_activation_time(cert);
         if(notValidBefore == (time_t)-1) {
             xmlSecGnuTLSError2("gnutls_x509_crt_get_activation_time", GNUTLS_E_SUCCESS,
-                               NULL,
-                               "cert activation time is invalid: %ld",
-                               (unsigned long)notValidBefore);
+                NULL,
+                "cert activation time is invalid: %.lf",
+                difftime(notValidBefore, (time_t)0));
             return(-1);
         }
         notValidAfter = gnutls_x509_crt_get_expiration_time(cert);
         if(notValidAfter == (time_t)-1) {
             xmlSecGnuTLSError2("gnutls_x509_crt_get_expiration_time", GNUTLS_E_SUCCESS,
-                               NULL,
-                               "cert expiration time is invalid: %ld",
-                               (unsigned long)notValidAfter);
+                NULL,
+                "cert expiration time is invalid: %.lf",
+                difftime(notValidAfter, (time_t)0));
             return(-1);
        }
 
@@ -214,11 +265,11 @@ xmlSecGnuTLSX509StoreVerify(xmlSecKeyDataStorePtr store,
     gnutls_x509_crt_t res = NULL;
     xmlSecSize certs_size = 0;
     gnutls_x509_crt_t * cert_list = NULL;
-    xmlSecSize cert_list_length;
+    xmlSecSize cert_list_size;
     gnutls_x509_crl_t * crl_list = NULL;
-    xmlSecSize crl_list_length;
+    xmlSecSize crl_list_size;
     gnutls_x509_crt_t * ca_list = NULL;
-    xmlSecSize ca_list_length;
+    xmlSecSize ca_list_size;
     time_t verification_time;
     unsigned int flags = 0;
     xmlSecSize ii;
@@ -240,24 +291,24 @@ xmlSecGnuTLSX509StoreVerify(xmlSecKeyDataStorePtr store,
     xmlSecAssert2(ctx != NULL, NULL);
 
     /* Prepare */
-    cert_list_length = certs_size + xmlSecPtrListGetSize(&(ctx->certsUntrusted));
-    if(cert_list_length > 0) {
-        cert_list = (gnutls_x509_crt_t *)xmlMalloc(sizeof(gnutls_x509_crt_t) * cert_list_length);
+    cert_list_size = certs_size + xmlSecPtrListGetSize(&(ctx->certsUntrusted));
+    if(cert_list_size > 0) {
+        cert_list = (gnutls_x509_crt_t *)xmlMalloc(sizeof(gnutls_x509_crt_t) * cert_list_size);
         if(cert_list == NULL) {
-            xmlSecMallocError(sizeof(gnutls_x509_crt_t) * cert_list_length,
+            xmlSecMallocError(sizeof(gnutls_x509_crt_t) * cert_list_size,
                               xmlSecKeyDataStoreGetName(store));
             goto done;
         }
     }
-    crl_list_length = xmlSecPtrListGetSize(crls);
-    if(crl_list_length > 0) {
-        crl_list = (gnutls_x509_crl_t *)xmlMalloc(sizeof(gnutls_x509_crl_t) * crl_list_length);
+    crl_list_size = xmlSecPtrListGetSize(crls);
+    if(crl_list_size > 0) {
+        crl_list = (gnutls_x509_crl_t *)xmlMalloc(sizeof(gnutls_x509_crl_t) * crl_list_size);
         if(crl_list == NULL) {
-            xmlSecMallocError(sizeof(gnutls_x509_crl_t) * crl_list_length,
+            xmlSecMallocError(sizeof(gnutls_x509_crl_t) * crl_list_size,
                               xmlSecKeyDataStoreGetName(store));
             goto done;
         }
-        for(ii = 0; ii < crl_list_length; ++ii) {
+        for(ii = 0; ii < crl_list_size; ++ii) {
             crl_list[ii] = xmlSecPtrListGetItem(crls, ii);
             if(crl_list[ii] == NULL) {
                 xmlSecInternalError("xmlSecPtrListGetItem(crls)",
@@ -267,15 +318,15 @@ xmlSecGnuTLSX509StoreVerify(xmlSecKeyDataStorePtr store,
         }
     }
 
-    ca_list_length = xmlSecPtrListGetSize(&(ctx->certsTrusted));
-    if(ca_list_length > 0) {
-        ca_list = (gnutls_x509_crt_t *)xmlMalloc(sizeof(gnutls_x509_crt_t) * ca_list_length);
+    ca_list_size = xmlSecPtrListGetSize(&(ctx->certsTrusted));
+    if(ca_list_size > 0) {
+        ca_list = (gnutls_x509_crt_t *)xmlMalloc(sizeof(gnutls_x509_crt_t) * ca_list_size);
         if(ca_list == NULL) {
-            xmlSecMallocError(sizeof(gnutls_x509_crt_t) * ca_list_length,
+            xmlSecMallocError(sizeof(gnutls_x509_crt_t) * ca_list_size,
                               xmlSecKeyDataStoreGetName(store));
             goto done;
         }
-        for(ii = 0; ii < ca_list_length; ++ii) {
+        for(ii = 0; ii < ca_list_size; ++ii) {
             ca_list[ii] = xmlSecPtrListGetItem(&(ctx->certsTrusted), ii);
             if(ca_list[ii] == NULL) {
                 xmlSecInternalError("xmlSecPtrListGetItem(certsTrusted)",
@@ -295,12 +346,15 @@ xmlSecGnuTLSX509StoreVerify(xmlSecKeyDataStorePtr store,
     if((keyInfoCtx->flags & XMLSEC_KEYINFO_FLAGS_X509DATA_SKIP_STRICT_CHECKS) != 0) {
         flags |= GNUTLS_VERIFY_ALLOW_SIGN_RSA_MD2;
         flags |= GNUTLS_VERIFY_ALLOW_SIGN_RSA_MD5;
+#if GNUTLS_VERSION_NUMBER >= 0x030600
+        flags |= GNUTLS_VERIFY_ALLOW_SIGN_WITH_SHA1;
+#endif
     }
 
     /* We are going to build all possible cert chains and try to verify them */
     for(ii = 0; (ii < certs_size) && (res == NULL); ++ii) {
         gnutls_x509_crt_t cert, cert2;
-        xmlSecSize cert_list_cur_length = 0;
+        xmlSecSize cert_list_cur_size = 0;
         unsigned int verify = 0;
 
         cert = xmlSecPtrListGetItem(certs, ii);
@@ -316,14 +370,14 @@ xmlSecGnuTLSX509StoreVerify(xmlSecKeyDataStorePtr store,
         }
 
         /* build the chain */
-        for(cert2 = cert, cert_list_cur_length = 0;
-            (cert2 != NULL) && (cert_list_cur_length < cert_list_length);
-            ++cert_list_cur_length)
+        for(cert2 = cert, cert_list_cur_size = 0;
+            (cert2 != NULL) && (cert_list_cur_size < cert_list_size);
+            ++cert_list_cur_size)
         {
             gnutls_x509_crt_t tmp;
 
             /* store */
-            cert_list[cert_list_cur_length] = cert2;
+            cert_list[cert_list_cur_size] = cert2;
 
             /* find next */
             tmp = xmlSecGnuTLSX509FindSignerCert(certs, cert2);
@@ -334,11 +388,17 @@ xmlSecGnuTLSX509StoreVerify(xmlSecKeyDataStorePtr store,
         }
 
         /* try to verify */
-	if((keyInfoCtx->flags & XMLSEC_KEYINFO_FLAGS_X509DATA_DONT_VERIFY_CERTS) == 0) {
+        if((keyInfoCtx->flags & XMLSEC_KEYINFO_FLAGS_X509DATA_DONT_VERIFY_CERTS) == 0) {
+            unsigned int cert_list_cur_len, ca_list_len, crl_list_len;
+
+            XMLSEC_SAFE_CAST_SIZE_TO_UINT(cert_list_cur_size, cert_list_cur_len, goto done, NULL);
+            XMLSEC_SAFE_CAST_SIZE_TO_UINT(ca_list_size, ca_list_len, goto done, NULL);
+            XMLSEC_SAFE_CAST_SIZE_TO_UINT(crl_list_size, crl_list_len, goto done, NULL);
+
             err = gnutls_x509_crt_list_verify(
-                    cert_list, (int)cert_list_cur_length, /* certs chain */
-                    ca_list, (int)ca_list_length, /* trusted cas */
-                    crl_list, (int)crl_list_length, /* crls */
+                    cert_list, cert_list_cur_len, /* certs chain */
+                    ca_list, ca_list_len, /* trusted cas */
+                    crl_list, crl_list_len, /* crls */
                     flags, /* flags */
                     &verify);
         } else {
@@ -350,15 +410,14 @@ xmlSecGnuTLSX509StoreVerify(xmlSecKeyDataStorePtr store,
             continue;
         } else if(verify != 0) {
             xmlSecOtherError2(XMLSEC_ERRORS_R_CERT_VERIFY_FAILED, NULL,
-                              "gnutls_x509_crt_list_verify: verification failed: status=%du",
-                              verify);
+                "gnutls_x509_crt_list_verify: verification failed: status=%u", verify);
             /* don't stop, continue! */
             continue;
         }
 
         /* gnutls doesn't allow to specify "verification" timestamp so
            we have to do it ourselves */
-        ret = xmlSecGnuTLSX509CheckTime(cert_list, cert_list_cur_length, verification_time);
+        ret = xmlSecGnuTLSX509CheckTime(cert_list, cert_list_cur_size, verification_time);
         if(ret != 1) {
             xmlSecInternalError("xmlSecGnuTLSX509CheckTime", NULL);
             /* don't stop, continue! */
@@ -526,12 +585,75 @@ done:
     return(res);
 }
 
+
+/**
+ * xmlSecGnuTLSX509CertCompareSKI:
+ *
+ * Returns 0 if SKI matches, 1 if SKI doesn't match and a negative value if an error occurs.
+ */
+static int
+xmlSecGnuTLSX509CertCompareSKI(gnutls_x509_crt_t cert, const xmlSecByte * ski, xmlSecSize skiSize) {
+    xmlSecByte* buf = NULL;
+    size_t bufSizeT = 0;
+    xmlSecSize bufSize;
+    unsigned int critical = 0;
+    int err;
+    int res = -1;
+
+    xmlSecAssert2(cert != NULL, -1);
+    xmlSecAssert2(ski != NULL, -1);
+    xmlSecAssert2(skiSize > 0, -1);
+
+    /* get ski size */
+    err = gnutls_x509_crt_get_subject_key_id(cert, NULL, &bufSizeT, &critical);
+    if((err != GNUTLS_E_SHORT_MEMORY_BUFFER) || (bufSizeT <= 0)) {
+        xmlSecGnuTLSError("gnutls_x509_crt_get_subject_key_id", err, NULL);
+        goto done;
+    }
+    XMLSEC_SAFE_CAST_SIZE_T_TO_SIZE(bufSizeT, bufSize, goto done, NULL);
+
+    if(skiSize != bufSize) {
+        /* doesn't match */
+        res = 1;
+        goto done;
+    }
+
+    /* allocate buffer */
+    buf = (xmlSecByte *)xmlMalloc(bufSizeT + 1);
+    if(buf == NULL) {
+        xmlSecMallocError(bufSizeT + 1, NULL);
+        goto done;
+    }
+
+    /* write ski out */
+    err = gnutls_x509_crt_get_subject_key_id(cert, buf, &bufSizeT, &critical);
+    if(err != GNUTLS_E_SUCCESS) {
+        xmlSecGnuTLSError("gnutls_x509_crt_get_subject_key_id", err, NULL);
+        goto done;
+    }
+
+    /* compare */
+    if(memcmp(ski, buf, bufSize) != 0) {
+        /* doesn't match */
+        res = 1;
+        goto done;
+    }
+
+    /* match! */
+    res = 0;
+
+done:
+    /* cleanup */
+    if(buf != NULL) {
+        xmlFree(buf);
+    }
+    return(res);
+}
+
 static gnutls_x509_crt_t
-xmlSecGnuTLSX509FindCert(xmlSecPtrListPtr certs,
-                         const xmlChar *subjectName,
-                         const xmlChar *issuerName,
-                         const xmlChar *issuerSerial,
-                         const xmlChar *ski) {
+xmlSecGnuTLSX509FindCert(xmlSecPtrListPtr certs, const xmlChar *subjectName,
+                         const xmlChar *issuerName, const xmlChar *issuerSerial,
+                         const xmlSecByte * ski, xmlSecSize skiSize) {
     xmlSecSize ii, sz;
 
     xmlSecAssert2(certs != NULL, NULL);
@@ -542,17 +664,18 @@ xmlSecGnuTLSX509FindCert(xmlSecPtrListPtr certs,
         gnutls_x509_crt_t cert = xmlSecPtrListGetItem(certs, ii);
         if(cert == NULL) {
             xmlSecInternalError2("xmlSecPtrListGetItem", NULL,
-                                 "pos=%i", (int)ii);
+                "pos=" XMLSEC_SIZE_FMT, ii);
             return(NULL);
         }
 
+        /* check subject name */
         if(subjectName != NULL) {
             xmlChar * tmp;
 
             tmp = xmlSecGnuTLSX509CertGetSubjectDN(cert);
             if(tmp == NULL) {
                 xmlSecInternalError2("xmlSecGnuTLSX509CertGetSubjectDN", NULL,
-                                     "pos=%i", (int)ii);
+                    "pos=" XMLSEC_SIZE_FMT, ii);
                 return(NULL);
             }
 
@@ -561,21 +684,24 @@ xmlSecGnuTLSX509FindCert(xmlSecPtrListPtr certs,
                 return(cert);
             }
             xmlFree(tmp);
-        } else if((issuerName != NULL) && (issuerSerial != NULL)) {
+        }
+
+        /* check issuer name + serial */
+        if((issuerName != NULL) && (issuerSerial != NULL)) {
             xmlChar * tmp1;
             xmlChar * tmp2;
 
             tmp1 = xmlSecGnuTLSX509CertGetIssuerDN(cert);
             if(tmp1 == NULL) {
                 xmlSecInternalError2("xmlSecGnuTLSX509CertGetIssuerDN", NULL,
-                                     "pos=%i", (int)ii);
+                    "pos=" XMLSEC_SIZE_FMT, ii);
                 return(NULL);
             }
 
             tmp2 = xmlSecGnuTLSX509CertGetIssuerSerial(cert);
             if(tmp2 == NULL) {
                 xmlSecInternalError2("xmlSecGnuTLSX509CertGetIssuerSerial", NULL,
-                                     "pos=%i", (int)ii);
+                    "pos=" XMLSEC_SIZE_FMT, ii);
                 xmlFree(tmp1);
                 return(NULL);
             }
@@ -587,21 +713,21 @@ xmlSecGnuTLSX509FindCert(xmlSecPtrListPtr certs,
             }
             xmlFree(tmp1);
             xmlFree(tmp2);
-        } else if(ski != NULL) {
-            xmlChar * tmp;
+        }
 
-            tmp = xmlSecGnuTLSX509CertGetSKI(cert);
-            if(tmp == NULL) {
-                xmlSecInternalError2("xmlSecGnuTLSX509CertGetSKI", NULL,
-                                     "pos=%i", (int)ii);
+        /* check subject ski */
+        if((ski != NULL) && (skiSize > 0)) {
+            int ret;
+
+            ret = xmlSecGnuTLSX509CertCompareSKI(cert, ski, skiSize);
+            if(ret < 0) {
+                xmlSecInternalError2("xmlSecGnuTLSX509CertCompareSKI", NULL,
+                    "pos=" XMLSEC_SIZE_FMT, ii);
                 return(NULL);
             }
-
-            if(xmlStrEqual(ski, tmp)) {
-                xmlFree(tmp);
+            if(ret == 0) {
                 return(cert);
             }
-            xmlFree(tmp);
         }
     }
 
@@ -634,14 +760,14 @@ xmlSecGnuTLSX509FindSignedCert(xmlSecPtrListPtr certs, gnutls_x509_crt_t cert) {
         tmp = xmlSecPtrListGetItem(certs, ii);
         if(tmp == NULL) {
             xmlSecInternalError2("xmlSecPtrListGetItem", NULL,
-                                 "pos=%i", (int)ii);
+                "pos=" XMLSEC_SIZE_FMT, ii);
             goto done;
         }
 
         issuer = xmlSecGnuTLSX509CertGetIssuerDN(tmp);
         if(issuer == NULL) {
             xmlSecInternalError2("xmlSecGnuTLSX509CertGetIssuerDN", NULL,
-                                 "pos=%i", (int)ii);
+                "pos=" XMLSEC_SIZE_FMT, ii);
             goto done;
         }
 
@@ -685,14 +811,14 @@ xmlSecGnuTLSX509FindSignerCert(xmlSecPtrListPtr certs, gnutls_x509_crt_t cert) {
         tmp = xmlSecPtrListGetItem(certs, ii);
         if(tmp == NULL) {
             xmlSecInternalError2("xmlSecPtrListGetItem", NULL,
-                                 "pos=%i", (int)ii);
+                "pos=" XMLSEC_SIZE_FMT, ii);
             goto done;
         }
 
         subject = xmlSecGnuTLSX509CertGetSubjectDN(tmp);
         if(subject == NULL) {
             xmlSecInternalError2("xmlSecGnuTLSX509CertGetSubjectDN", NULL,
-                                 "pos=%i", (int)ii);
+                "pos=" XMLSEC_SIZE_FMT, ii);
             goto done;
         }
 
