@@ -9,7 +9,7 @@
  */
 /**
  * SECTION:x509
- * @Short_description: X509 certificates implementation for Microsoft Cryptography API: Next Generation (CNG). 
+ * @Short_description: X509 certificates implementation for Microsoft Cryptography API: Next Generation (CNG).
  * @Stability: Stable
  *
  */
@@ -23,18 +23,21 @@
 #include <windows.h>
 
 #include <xmlsec/xmlsec.h>
-#include <xmlsec/xmltree.h>
-#include <xmlsec/keys.h>
-#include <xmlsec/keyinfo.h>
-#include <xmlsec/keysmngr.h>
-#include <xmlsec/x509.h>
 #include <xmlsec/base64.h>
 #include <xmlsec/bn.h>
 #include <xmlsec/errors.h>
+#include <xmlsec/keys.h>
+#include <xmlsec/keyinfo.h>
+#include <xmlsec/keysmngr.h>
+#include <xmlsec/xmltree.h>
+#include <xmlsec/x509.h>
 
+#include <xmlsec/mscng/certkeys.h>
 #include <xmlsec/mscng/crypto.h>
 #include <xmlsec/mscng/x509.h>
-#include <xmlsec/mscng/certkeys.h>
+
+#include "../cast_helpers.h"
+#include "../keysdata_helpers.h"
 
 typedef struct _xmlSecMSCngX509DataCtx xmlSecMSCngX509DataCtx,
                                        *xmlSecMSCngX509DataCtxPtr;
@@ -44,10 +47,8 @@ struct _xmlSecMSCngX509DataCtx {
     PCCERT_CONTEXT cert;
 };
 
-#define xmlSecMSCngX509DataSize      \
-    (sizeof(xmlSecKeyData) + sizeof(xmlSecMSCngX509DataCtx))
-#define xmlSecMSCngX509DataGetCtx(data) \
-    ((xmlSecMSCngX509DataCtxPtr)(((xmlSecByte*)(data)) + sizeof(xmlSecKeyData)))
+XMLSEC_KEY_DATA_DECLARE(MSCngX509Data, xmlSecMSCngX509DataCtx)
+#define xmlSecMSCngX509DataSize XMLSEC_KEY_DATA_SIZE(MSCngX509Data)
 
 static int
 xmlSecMSCngKeyDataX509Initialize(xmlSecKeyDataPtr data) {
@@ -167,11 +168,13 @@ xmlSecMSCngKeyDataX509GetIdentifier(xmlSecKeyDataPtr data) {
 static PCCERT_CONTEXT
 xmlSecMSCngX509CertDerRead(const xmlSecByte* buf, xmlSecSize size) {
     PCCERT_CONTEXT cert;
+    DWORD dwSize;
 
     xmlSecAssert2(buf != NULL, NULL);
     xmlSecAssert2(size > 0, NULL);
 
-    cert = CertCreateCertificateContext(X509_ASN_ENCODING | PKCS_7_ASN_ENCODING, buf, size);
+    XMLSEC_SAFE_CAST_SIZE_TO_ULONG(size, dwSize, return(NULL), NULL);
+    cert = CertCreateCertificateContext(X509_ASN_ENCODING | PKCS_7_ASN_ENCODING, buf, dwSize);
     if(cert == NULL) {
         xmlSecMSCngLastError("CertCreateCertificateContext", NULL);
         return(NULL);
@@ -180,28 +183,6 @@ xmlSecMSCngX509CertDerRead(const xmlSecByte* buf, xmlSecSize size) {
     return(cert);
 }
 
-/**
- * xmlSecMSCngX509CertBase64DerRead:
- *
- * The MSCng reader for the <X509Certificate> XML content.
- */
-static PCCERT_CONTEXT
-xmlSecMSCngX509CertBase64DerRead(xmlChar* buf) {
-    int size;
-
-    xmlSecAssert2(buf != NULL, NULL);
-
-    /* in-place decoding */
-    size = xmlSecBase64Decode(buf, (xmlSecByte*)buf, xmlStrlen(buf));
-    if(size < 0) {
-        xmlSecInternalError("xmlSecBase64Decode", NULL);
-        return(NULL);
-    }
-
-    return(xmlSecMSCngX509CertDerRead((xmlSecByte*)buf, size));
-}
-
-
 int
 xmlSecMSCngKeyDataX509AdoptKeyCert(xmlSecKeyDataPtr data, PCCERT_CONTEXT cert) {
     xmlSecMSCngX509DataCtxPtr ctx;
@@ -274,315 +255,17 @@ xmlSecMSCngKeyDataX509AdoptCrl(xmlSecKeyDataPtr data, PCCRL_CONTEXT crl) {
     return(0);
 }
 
-/**
- * xmlSecMSCngX509SubjectNameNodeRead:
- *
- * The MSCng reader for the <X509SubjectName> XML element.
- */
-static int
-xmlSecMSCngX509SubjectNameNodeRead(xmlSecKeyDataPtr data, xmlNodePtr node,
-    xmlSecKeyInfoCtxPtr keyInfoCtx) {
-    xmlSecKeyDataStorePtr store;
-    xmlChar* subject;
-    PCCERT_CONTEXT cert;
-    int ret;
-
-    xmlSecAssert2(xmlSecKeyDataCheckId(data, xmlSecMSCngKeyDataX509Id), -1);
-    xmlSecAssert2(node != NULL, -1);
-    xmlSecAssert2(keyInfoCtx != NULL, -1);
-    xmlSecAssert2(keyInfoCtx->keysMngr != NULL, -1);
-
-    store = xmlSecKeysMngrGetDataStore(keyInfoCtx->keysMngr, xmlSecMSCngX509StoreId);
-    if(store == NULL) {
-        xmlSecInternalError("xmlSecKeysMngrGetDataStore",
-            xmlSecKeyDataGetName(data));
-        return(-1);
-    }
-
-    subject = xmlNodeGetContent(node);
-    if((subject == NULL) || (xmlSecIsEmptyString(subject) == 1)) {
-        if(subject != NULL) {
-            xmlFree(subject);
-        }
-        if((keyInfoCtx->flags & XMLSEC_KEYINFO_FLAGS_STOP_ON_EMPTY_NODE) != 0) {
-            xmlSecInvalidNodeContentError(node, xmlSecKeyDataGetName(data),
-                "empty");
-            return(-1);
-        }
-
-        return(0);
-    }
-
-    cert = xmlSecMSCngX509StoreFindCert(store, subject, NULL, NULL, NULL, keyInfoCtx);
-    if(cert == NULL) {
-        if((keyInfoCtx->flags & XMLSEC_KEYINFO_FLAGS_X509DATA_STOP_ON_UNKNOWN_CERT) != 0) {
-            xmlSecOtherError2(XMLSEC_ERRORS_R_CERT_NOT_FOUND,
-                xmlSecKeyDataGetName(data), "subject=%s",
-                xmlSecErrorsSafeString(subject));
-            xmlFree(subject);
-            return(-1);
-        }
-
-        xmlFree(subject);
-        return(0);
-    }
-
-    ret = xmlSecMSCngKeyDataX509AdoptCert(data, cert);
-    if(ret < 0) {
-        xmlSecInternalError("xmlSecMSCngKeyDataX509AdoptCert",
-            xmlSecKeyDataGetName(data));
-        CertFreeCertificateContext(cert);
-        xmlFree(subject);
-        return(-1);
-    }
-
-    xmlFree(subject);
-    return(0);
-}
-/**
- * xmlSecMSCngX509CertificateNodeRead:
- *
- * The MSCng reader for the <X509Certificate> XML element.
- */
-static int
-xmlSecMSCngX509CertificateNodeRead(xmlSecKeyDataPtr data, xmlNodePtr node,
-    xmlSecKeyInfoCtxPtr keyInfoCtx) {
-    xmlChar* content;
-    PCCERT_CONTEXT cert;
-    int ret;
-
-    xmlSecAssert2(xmlSecKeyDataCheckId(data, xmlSecMSCngKeyDataX509Id), -1);
-    xmlSecAssert2(node != NULL, -1);
-    xmlSecAssert2(keyInfoCtx != NULL, -1);
-
-    content = xmlNodeGetContent(node);
-    if((content == NULL) || (xmlSecIsEmptyString(content) == 1)) {
-        if(content != NULL) {
-            xmlFree(content);
-        }
-
-        if((keyInfoCtx->flags & XMLSEC_KEYINFO_FLAGS_STOP_ON_EMPTY_NODE) != 0) {
-            xmlSecInvalidNodeContentError(node, xmlSecKeyDataGetName(data),
-                "content is an empty string");
-            return(-1);
-        }
-
-        return(0);
-    }
-
-    cert = xmlSecMSCngX509CertBase64DerRead(content);
-    if(cert == NULL) {
-        xmlSecInternalError("xmlSecMSCngX509CertBase64DerRead",
-            xmlSecKeyDataGetName(data));
-        return(-1);
-    }
-
-    ret = xmlSecMSCngKeyDataX509AdoptCert(data, cert);
-    if(ret < 0) {
-        xmlSecInternalError("xmlSecMSCngKeyDataX509AdoptCert",
-            xmlSecKeyDataGetName(data));
-        return(-1);
-
-    }
-
-    xmlFree(content);
-    return(0);
-}
-
-/**
- * xmlSecMSCngX509IssuerSerialNodeRead:
- *
- * The MSCng reader for the <X509IssuerSerial> XML element.
- */
-static int
-xmlSecMSCngX509IssuerSerialNodeRead(xmlSecKeyDataPtr data, xmlNodePtr node,
-    xmlSecKeyInfoCtxPtr keyInfoCtx) {
-    xmlSecKeyDataStorePtr store;
-    xmlNodePtr cur;
-    xmlChar* issuerName;
-    xmlChar* issuerSerial;
-    PCCERT_CONTEXT cert;
-    int ret;
-
-    xmlSecAssert2(xmlSecKeyDataCheckId(data, xmlSecMSCngKeyDataX509Id), -1);
-    xmlSecAssert2(node != NULL, -1);
-    xmlSecAssert2(keyInfoCtx != NULL, -1);
-    xmlSecAssert2(keyInfoCtx->keysMngr != NULL, -1);
-
-    store = xmlSecKeysMngrGetDataStore(keyInfoCtx->keysMngr, xmlSecMSCngX509StoreId);
-    if(store == NULL) {
-        xmlSecInternalError("xmlSecKeysMngrGetDataStore",
-            xmlSecKeyDataGetName(data));
-        return(-1);
-    }
-
-    cur = xmlSecGetNextElementNode(node->children);
-    if(cur == NULL) {
-        if((keyInfoCtx->flags & XMLSEC_KEYINFO_FLAGS_STOP_ON_EMPTY_NODE) != 0) {
-            xmlSecNodeNotFoundError("xmlSecGetNextElementNode", node, NULL,
-                xmlSecKeyDataGetName(data));
-            return(-1);
-        }
-
-        return(0);
-    }
-
-    /* handle X509IssuerName */
-    if(!xmlSecCheckNodeName(cur, xmlSecNodeX509IssuerName, xmlSecDSigNs)) {
-        xmlSecInvalidNodeError(cur, xmlSecNodeX509IssuerName,
-            xmlSecKeyDataGetName(data));
-        return(-1);
-    }
-
-    issuerName = xmlNodeGetContent(cur);
-    if(issuerName == NULL) {
-        xmlSecInvalidNodeContentError(cur, xmlSecKeyDataGetName(data),
-            "empty");
-        return(-1);
-    }
-
-    cur = xmlSecGetNextElementNode(cur->next);
-    if(cur == NULL) {
-        if((keyInfoCtx->flags & XMLSEC_KEYINFO_FLAGS_STOP_ON_EMPTY_NODE) != 0) {
-            xmlSecNodeNotFoundError("xmlSecGetNextElementNode", node, NULL,
-                xmlSecKeyDataGetName(data));
-            return(-1);
-        }
-
-        return(0);
-    }
-
-    /* handle X509SerialNumber */
-    if(!xmlSecCheckNodeName(cur, xmlSecNodeX509SerialNumber, xmlSecDSigNs)) {
-        xmlSecInvalidNodeError(cur, xmlSecNodeX509SerialNumber,
-            xmlSecKeyDataGetName(data));
-        xmlFree(issuerName);
-        return(-1);
-    }
-
-    issuerSerial = xmlNodeGetContent(cur);
-    if(issuerSerial == NULL) {
-        xmlSecInvalidNodeContentError(cur, xmlSecKeyDataGetName(data),
-            "empty");
-        xmlFree(issuerSerial);
-        return(-1);
-    }
-
-    cur = xmlSecGetNextElementNode(cur->next);
-    if(cur != NULL) {
-        xmlSecUnexpectedNodeError(cur, xmlSecKeyDataGetName(data));
-        xmlFree(issuerSerial);
-        xmlFree(issuerName);
-        return(-1);
-    }
-
-    cert = xmlSecMSCngX509StoreFindCert(store, NULL, issuerName, issuerSerial,
-        NULL, keyInfoCtx);
-    if(cert == NULL) {
-        if((keyInfoCtx->flags & XMLSEC_KEYINFO_FLAGS_X509DATA_STOP_ON_UNKNOWN_CERT) != 0) {
-            xmlSecOtherError3(XMLSEC_ERRORS_R_CERT_NOT_FOUND, xmlSecKeyDataGetName(data),
-                "issuerName=%s;issuerSerial=%s",
-                xmlSecErrorsSafeString(issuerName),
-                xmlSecErrorsSafeString(issuerSerial));
-            xmlFree(issuerSerial);
-            xmlFree(issuerName);
-            return(-1);
-        }
-
-	xmlFree(issuerSerial);
-	xmlFree(issuerName);
-	return(0);
-    }
-
-    ret = xmlSecMSCngKeyDataX509AdoptCert(data, cert);
-    if(ret < 0) {
-        xmlSecInternalError("xmlSecMSCngKeyDataX509AdoptCert",
-            xmlSecKeyDataGetName(data));
-        CertFreeCertificateContext(cert);
-        xmlFree(issuerSerial);
-        xmlFree(issuerName);
-        return(-1);
-    }
-
-    xmlFree(issuerSerial);
-    xmlFree(issuerName);
-    return(0);
-}
-
-/**
- * xmlSecMSCngX509SKINodeRead:
- *
- * The MSCng reader for the <X509SKI> XML element.
- */
-static int
-xmlSecMSCngX509SKINodeRead(xmlSecKeyDataPtr data, xmlNodePtr node,
-        xmlSecKeyInfoCtxPtr keyInfoCtx) {
-    xmlSecKeyDataStorePtr store;
-    xmlChar* ski;
-    PCCERT_CONTEXT cert;
-    int ret;
-
-    xmlSecAssert2(xmlSecKeyDataCheckId(data, xmlSecMSCngKeyDataX509Id), -1);
-    xmlSecAssert2(node != NULL, -1);
-    xmlSecAssert2(keyInfoCtx != NULL, -1);
-    xmlSecAssert2(keyInfoCtx->keysMngr != NULL, -1);
-
-    store = xmlSecKeysMngrGetDataStore(keyInfoCtx->keysMngr, xmlSecMSCngX509StoreId);
-    if(store == NULL) {
-        xmlSecInternalError("xmlSecKeysMngrGetDataStore",
-            xmlSecKeyDataGetName(data));
-        return(-1);
-    }
-
-    ski = xmlNodeGetContent(node);
-    if((ski == NULL) || (xmlSecIsEmptyString(ski) == 1)) {
-        if(ski != NULL) {
-            xmlFree(ski);
-        }
-        if((keyInfoCtx->flags & XMLSEC_KEYINFO_FLAGS_STOP_ON_EMPTY_NODE) != 0) {
-            xmlSecInvalidNodeContentError(node, xmlSecKeyDataGetName(data),
-                "empty");
-            return(-1);
-        }
-        return(0);
-    }
-
-    cert = xmlSecMSCngX509StoreFindCert(store, NULL, NULL, NULL, ski, keyInfoCtx);
-    if(cert == NULL){
-        xmlFree(ski);
-
-        if((keyInfoCtx->flags & XMLSEC_KEYINFO_FLAGS_X509DATA_STOP_ON_UNKNOWN_CERT) != 0) {
-            xmlSecOtherError2(XMLSEC_ERRORS_R_CERT_NOT_FOUND, xmlSecKeyDataGetName(data),
-                              "ski=%s", xmlSecErrorsSafeString(ski));
-            return(-1);
-        }
-        return(0);
-    }
-
-    ret = xmlSecMSCngKeyDataX509AdoptCert(data, cert);
-    if(ret < 0) {
-        xmlSecInternalError("xmlSecMSCngKeyDataX509AdoptCert",
-                            xmlSecKeyDataGetName(data));
-        CertFreeCertificateContext(cert);
-        xmlFree(ski);
-        return(-1);
-    }
-
-    xmlFree(ski);
-    return(0);
-}
 
 static PCCRL_CONTEXT
-xmlSecMSCngX509CrlDerRead(xmlSecByte* buf, xmlSecSize size,
-        xmlSecKeyInfoCtxPtr keyInfoCtx) {
+xmlSecMSCngX509CrlDerRead(xmlSecByte* buf, xmlSecSize size) {
     PCCRL_CONTEXT crl = NULL;
+    DWORD dwSize;
 
     xmlSecAssert2(buf != NULL, NULL);
-    xmlSecAssert2(keyInfoCtx != NULL, NULL);
     xmlSecAssert2(size > 0, NULL);
 
-    crl = CertCreateCRLContext(X509_ASN_ENCODING | PKCS_7_ASN_ENCODING, buf, size);
-
+    XMLSEC_SAFE_CAST_SIZE_TO_ULONG(size, dwSize, return(NULL), NULL);
+    crl = CertCreateCRLContext(X509_ASN_ENCODING | PKCS_7_ASN_ENCODING, buf, dwSize);
     if(crl == NULL) {
         xmlSecMSCngLastError("CertCreateCRLContext", NULL);
         return(NULL);
@@ -591,124 +274,6 @@ xmlSecMSCngX509CrlDerRead(xmlSecByte* buf, xmlSecSize size,
     return(crl);
 }
 
-static PCCRL_CONTEXT
-xmlSecMSCngX509CrlBase64DerRead(xmlChar* buf, xmlSecKeyInfoCtxPtr keyInfoCtx) {
-    int ret;
-
-    xmlSecAssert2(buf != NULL, NULL);
-
-    /* usual trick with base64 decoding in-place */
-    ret = xmlSecBase64Decode(buf, (xmlSecByte*)buf, xmlStrlen(buf));
-    if(ret < 0) {
-        xmlSecInternalError("xmlSecBase64Decode", NULL);
-        return(NULL);
-    }
-
-    return(xmlSecMSCngX509CrlDerRead((xmlSecByte*)buf, ret, keyInfoCtx));
-}
-
-static int
-xmlSecMSCngX509CRLNodeRead(xmlSecKeyDataPtr data, xmlNodePtr node, xmlSecKeyInfoCtxPtr keyInfoCtx) {
-    xmlChar *content;
-    PCCRL_CONTEXT crl;
-    int ret;
-
-    xmlSecAssert2(xmlSecKeyDataCheckId(data, xmlSecMSCngKeyDataX509Id), -1);
-    xmlSecAssert2(node != NULL, -1);
-    xmlSecAssert2(keyInfoCtx != NULL, -1);
-
-    content = xmlNodeGetContent(node);
-    if((content == NULL) || (xmlSecIsEmptyString(content) == 1)) {
-        if(content != NULL) {
-            xmlFree(content);
-        }
-        if((keyInfoCtx->flags & XMLSEC_KEYINFO_FLAGS_STOP_ON_EMPTY_NODE) != 0) {
-            xmlSecInvalidNodeContentError(node, xmlSecKeyDataGetName(data), "empty");
-            return(-1);
-        }
-        return(0);
-    }
-
-    crl = xmlSecMSCngX509CrlBase64DerRead(content, keyInfoCtx);
-    if(crl == NULL) {
-        xmlSecInternalError("xmlSecMSCngX509CrlBase64DerRead",
-            xmlSecKeyDataGetName(data));
-        xmlFree(content);
-        return(-1);
-    }
-
-    ret = xmlSecMSCngKeyDataX509AdoptCrl(data, crl);
-    if (ret < 0) {
-        xmlSecInternalError("xmlSecMSCngKeyDataX509AdoptCrl",
-            xmlSecKeyDataGetName(data));
-        xmlFree(content);
-        CertFreeCRLContext(crl);
-        return(-1);
-    }
-
-    xmlFree(content);
-    return(0);
-}
-
-/**
- * xmlSecMSCngX509DataNodeRead:
- *
- * The MSCng reader for the <X509Data> XML element.
- */
-static int
-xmlSecMSCngX509DataNodeRead(xmlSecKeyDataPtr data, xmlNodePtr node,
-    xmlSecKeyInfoCtxPtr keyInfoCtx) {
-    xmlNodePtr cur;
-    int ret;
-
-    xmlSecAssert2(xmlSecKeyDataCheckId(data, xmlSecMSCngKeyDataX509Id), -1);
-    xmlSecAssert2(node != NULL, -1);
-    xmlSecAssert2(keyInfoCtx != NULL, -1);
-
-    for(cur = xmlSecGetNextElementNode(node->children);
-        cur != NULL;
-        cur = xmlSecGetNextElementNode(cur->next)) {
-        if(xmlSecCheckNodeName(cur, xmlSecNodeX509Certificate, xmlSecDSigNs)) {
-            ret = xmlSecMSCngX509CertificateNodeRead(data, cur, keyInfoCtx);
-            if(ret < 0) {
-                xmlSecInternalError("xmlSecMSCngX509CertificateNodeRead",
-                    xmlSecKeyDataGetName(data));
-                return(-1);
-            }
-        } else if(xmlSecCheckNodeName(cur, xmlSecNodeX509SubjectName, xmlSecDSigNs)) {
-            ret = xmlSecMSCngX509SubjectNameNodeRead(data, cur, keyInfoCtx);
-            if(ret < 0) {
-                xmlSecInternalError("xmlSecMSCngX509SubjectNameNodeRead",
-                    xmlSecKeyDataGetName(data));
-                return(-1);
-            }
-        } else if(xmlSecCheckNodeName(cur, xmlSecNodeX509IssuerSerial, xmlSecDSigNs)) {
-            ret = xmlSecMSCngX509IssuerSerialNodeRead(data, cur, keyInfoCtx);
-            if(ret < 0) {
-                xmlSecInternalError("xmlSecMSCngX509IssuerSerialNodeRead", NULL);
-                return(-1);
-            }
-        } else if(xmlSecCheckNodeName(cur, xmlSecNodeX509SKI, xmlSecDSigNs)) {
-            ret = xmlSecMSCngX509SKINodeRead(data, cur, keyInfoCtx);
-            if(ret < 0) {
-                xmlSecInternalError("xmlSecMSCngX509SKINodeRead", NULL);
-                return(-1);
-            }
-        } else if(xmlSecCheckNodeName(cur, xmlSecNodeX509CRL, xmlSecDSigNs)) {
-            ret = xmlSecMSCngX509CRLNodeRead(data, cur, keyInfoCtx);
-            if(ret < 0) {
-                xmlSecInternalError("xmlSecMSCngX509CRLNodeRead",
-                                    xmlSecKeyDataGetName(data));
-                return(-1);
-            }
-        } else if((keyInfoCtx->flags & XMLSEC_KEYINFO_FLAGS_X509DATA_STOP_ON_UNKNOWN_CHILD) != 0) {
-            xmlSecUnexpectedNodeError(cur, xmlSecKeyDataGetName(data));
-            return(-1);
-        }
-    }
-    return(0);
-}
-
 /**
  * xmlSecMSCngX509CertGetTime:
  *
@@ -841,6 +406,112 @@ xmlSecMSCngKeyDataX509VerifyAndExtractKey(xmlSecKeyDataPtr data,
     return(0);
 }
 
+static int
+xmlSecMSCngKeyDataX509Read(xmlSecKeyDataPtr data, xmlSecKeyValueX509Ptr x509Value,
+    xmlSecKeysMngrPtr keysMngr, unsigned int flags) {
+    xmlSecKeyDataStorePtr x509Store;
+    int stopOnUnknownCert = 0;
+    PCCERT_CONTEXT cert = NULL;
+    PCCRL_CONTEXT crl = NULL;
+    int ret;
+    int res = -1;
+
+    xmlSecAssert2(data != NULL, -1);
+    xmlSecAssert2(xmlSecKeyDataCheckId(data, xmlSecMSCngKeyDataX509Id), -1);
+    xmlSecAssert2(x509Value != NULL, -1);
+    xmlSecAssert2(keysMngr != NULL, -1);
+
+    x509Store = xmlSecKeysMngrGetDataStore(keysMngr, xmlSecMSCngX509StoreId);
+    if (x509Store == NULL) {
+        xmlSecInternalError("xmlSecKeysMngrGetDataStore", xmlSecKeyDataGetName(data));
+        goto done;
+    }
+
+    /* determine what to do */
+    if ((flags & XMLSEC_KEYINFO_FLAGS_X509DATA_STOP_ON_UNKNOWN_CERT) != 0) {
+        stopOnUnknownCert = 1;
+    }
+
+    if (xmlSecBufferGetSize(&(x509Value->cert)) > 0) {
+        cert = xmlSecMSCngX509CertDerRead(xmlSecBufferGetData(&(x509Value->cert)),
+            xmlSecBufferGetSize(&(x509Value->cert)));
+        if (cert == NULL) {
+            xmlSecInternalError("xmlSecMSCngX509CertDerRead", xmlSecKeyDataGetName(data));
+            goto done;
+        }
+    }
+    else if (xmlSecBufferGetSize(&(x509Value->crl)) > 0) {
+        crl = xmlSecMSCngX509CrlDerRead(xmlSecBufferGetData(&(x509Value->crl)),
+            xmlSecBufferGetSize(&(x509Value->crl)));
+        if (crl == NULL) {
+            xmlSecInternalError("xmlSecMSCngX509CertDerRead", xmlSecKeyDataGetName(data));
+            goto done;
+        }
+    }
+    else if (xmlSecBufferGetSize(&(x509Value->ski)) > 0) {
+        cert = xmlSecMSCngX509StoreFindCert_ex(x509Store, NULL, NULL, NULL,
+            xmlSecBufferGetData(&(x509Value->ski)), xmlSecBufferGetSize(&(x509Value->ski)),
+            NULL /* unused */);
+        if ((cert == NULL) && (stopOnUnknownCert != 0)) {
+            xmlSecOtherError2(XMLSEC_ERRORS_R_CERT_NOT_FOUND, xmlSecKeyDataGetName(data),
+                "skiSize=" XMLSEC_SIZE_FMT, xmlSecBufferGetSize(&(x509Value->ski)));
+            goto done;
+        }
+    }
+    else if (x509Value->subject != NULL) {
+        cert = xmlSecMSCngX509StoreFindCert_ex(x509Store, x509Value->subject,
+            NULL, NULL, NULL, 0, NULL /* unused */);
+        if ((cert == NULL) && (stopOnUnknownCert != 0)) {
+            xmlSecOtherError2(XMLSEC_ERRORS_R_CERT_NOT_FOUND, xmlSecKeyDataGetName(data),
+                "subject=%s", xmlSecErrorsSafeString(x509Value->subject));
+            goto done;
+        }
+    }
+    else if ((x509Value->issuerName != NULL) && (x509Value->issuerSerial != NULL)) {
+        cert = xmlSecMSCngX509StoreFindCert_ex(x509Store, NULL,
+            x509Value->issuerName, x509Value->issuerSerial,
+            NULL, 0, NULL /* unused */);
+        if ((cert == NULL) && (stopOnUnknownCert != 0)) {
+            xmlSecOtherError3(XMLSEC_ERRORS_R_CERT_NOT_FOUND, xmlSecKeyDataGetName(data),
+                "issuerName=%s;issuerSerial=%s",
+                xmlSecErrorsSafeString(x509Value->issuerName),
+                xmlSecErrorsSafeString(x509Value->issuerSerial));
+            goto done;
+        }
+    }
+
+    /* if we found a cert or a crl, then add it to the data */
+    if (cert != NULL) {
+        ret = xmlSecMSCngKeyDataX509AdoptCert(data, cert);
+        if (ret < 0) {
+            xmlSecInternalError("xmlSecMSCngKeyDataX509AdoptCert", xmlSecKeyDataGetName(data));
+            goto done;
+        }
+        cert = NULL; /* owned by data now */
+    }
+    if (crl != NULL) {
+        ret = xmlSecMSCngKeyDataX509AdoptCrl(data, crl);
+        if (ret < 0) {
+            xmlSecInternalError("xmlSecMSCngKeyDataX509AdoptCrl", xmlSecKeyDataGetName(data));
+            goto done;
+        }
+        crl = NULL; /* owned by data now */
+    }
+
+    /* success */
+    res = 0;
+
+done:
+    /* cleanup */
+    if (cert != NULL) {
+        CertFreeCertificateContext(cert);
+    }
+    if (crl != NULL) {
+        CertFreeCRLContext(crl);
+    }
+    return(res);
+}
+
 static int
 xmlSecMSCngKeyDataX509XmlRead(xmlSecKeyDataId id, xmlSecKeyPtr key,
                               xmlNodePtr node, xmlSecKeyInfoCtxPtr keyInfoCtx) {
@@ -849,62 +520,28 @@ xmlSecMSCngKeyDataX509XmlRead(xmlSecKeyDataId id, xmlSecKeyPtr key,
 
     xmlSecAssert2(id == xmlSecMSCngKeyDataX509Id, -1);
     xmlSecAssert2(key != NULL, -1);
-    xmlSecAssert2(node != NULL, -1);
-    xmlSecAssert2(keyInfoCtx != NULL, -1);
 
     data = xmlSecKeyEnsureData(key, id);
-    if(data == NULL) {
+    if (data == NULL) {
         xmlSecInternalError("xmlSecKeyEnsureData",
-                            xmlSecKeyDataKlassGetName(id));
+            xmlSecKeyDataKlassGetName(id));
         return(-1);
     }
 
-    ret = xmlSecMSCngX509DataNodeRead(data, node, keyInfoCtx);
-    if(ret < 0) {
-        xmlSecInternalError("xmlSecMSCngX509DataNodeRead",
+    ret = xmlSecKeyDataX509XmlRead(data, node, keyInfoCtx,
+        xmlSecMSCngKeyDataX509Read);
+    if (ret < 0) {
+        xmlSecInternalError("xmlSecKeyDataX509XmlRead",
             xmlSecKeyDataKlassGetName(id));
         return(-1);
     }
 
     ret = xmlSecMSCngKeyDataX509VerifyAndExtractKey(data, key, keyInfoCtx);
-    if(ret < 0) {
+    if (ret < 0) {
         xmlSecInternalError("xmlSecMSCngKeyDataX509VerifyAndExtractKey",
             xmlSecKeyDataKlassGetName(id));
         return(-1);
     }
-
-    return(0);
-}
-
-static int
-xmlSecMSCngX509CertificateNodeWrite(PCCERT_CONTEXT cert, xmlNodePtr node,
-        xmlSecKeyInfoCtxPtr keyInfoCtx) {
-    xmlChar* buf;
-    xmlNodePtr child;
-
-    xmlSecAssert2(cert != NULL, -1);
-    xmlSecAssert2(cert->pbCertEncoded != NULL, -1);
-    xmlSecAssert2(cert->cbCertEncoded > 0, -1);
-    xmlSecAssert2(node != NULL, -1);
-    xmlSecAssert2(keyInfoCtx != NULL, -1);
-
-    buf = xmlSecBase64Encode(cert->pbCertEncoded, cert->cbCertEncoded,
-        keyInfoCtx->base64LineSize);
-    if(buf == NULL) {
-        xmlSecInternalError("xmlSecBase64Encode", NULL);
-        return(-1);
-    }
-
-    child = xmlSecEnsureEmptyChild(node, xmlSecNodeX509Certificate, xmlSecDSigNs);
-    if(child == NULL) {
-        xmlSecInternalError("xmlSecEnsureEmptyChild", NULL);
-        xmlFree(buf);
-        return(-1);
-    }
-
-    xmlNodeSetContent(child, buf);
-    xmlFree(buf);
-
     return(0);
 }
 
@@ -948,307 +585,232 @@ xmlSecMSCngX509NameWrite(PCERT_NAME_BLOB nm) {
     return(res);
 }
 
-static int
-xmlSecMSCngX509SubjectNameNodeWrite(PCCERT_CONTEXT cert, xmlNodePtr node) {
-    xmlChar* buf = NULL;
-    xmlNodePtr cur = NULL;
-    int ret;
-
-    xmlSecAssert2(cert != NULL, -1);
-    xmlSecAssert2(node != NULL, -1);
-
-    buf = xmlSecMSCngX509NameWrite(&(cert->pCertInfo->Subject));
-    if(buf == NULL) {
-        xmlSecInternalError("xmlSecMSCngX509NameWrite", NULL);
-        return(-1);
-    }
-
-    cur = xmlSecEnsureEmptyChild(node, xmlSecNodeX509SubjectName, xmlSecDSigNs);
-    if(cur == NULL) {
-        xmlSecInternalError("xmlSecEnsureEmptyChild", NULL);
-        xmlFree(buf);
-        return(-1);
-    }
-
-    ret = xmlSecNodeEncodeAndSetContent(cur, buf);
-    if(ret < 0) {
-        xmlSecInternalError("xmlSecNodeEncodeAndSetContent", NULL);
-        xmlFree(buf);
-        return(-1);
-    }
-
-    /* done */
-    xmlFree(buf);
-    return(0);
-}
-
-static int
-xmlSecMSCngASN1IntegerWrite(xmlNodePtr node, PCRYPT_INTEGER_BLOB num) {
+static xmlChar*
+xmlSecMSCngASN1IntegerWrite(PCRYPT_INTEGER_BLOB num) {
     xmlSecBn bn;
+    xmlChar* res;
     int ret;
 
-    xmlSecAssert2(node != NULL, -1);
-    xmlSecAssert2(num != NULL, -1);
+    xmlSecAssert2(num != NULL, NULL);
 
     ret = xmlSecBnInitialize(&bn, num->cbData + 1);
     if(ret < 0) {
-	xmlSecInternalError2("xmlSecBnInitialize", NULL, "size=%ld",
-	    num->cbData + 1);
-        return(-1);
+        xmlSecInternalError2("xmlSecBnInitialize", NULL, "size=%lu", num->cbData + 1);
+        return(NULL);
     }
 
     ret = xmlSecBnSetData(&bn, num->pbData, num->cbData);
     if(ret < 0) {
         xmlSecInternalError("xmlSecBnSetData", NULL);
         xmlSecBnFinalize(&bn);
-        return(-1);
+        return(NULL);
     }
 
     /* SerialNumber is little-endian, see <https://msdn.microsoft.com/en-us/library/windows/desktop/aa377200(v=vs.85).aspx>.
-     * xmldsig wants big-endian, so enable reversing */
-    ret = xmlSecBnSetNodeValue(&bn, node, xmlSecBnDec, 1, 0);
-    if(ret < 0) {
-        xmlSecInternalError("xmlSecBnSetNodeValue", NULL);
+     * xmldsig wants big-endian, so reverse */
+    ret = xmlSecBnReverse(&bn);
+    if (ret < 0) {
+        xmlSecInternalError("xmlSecBnReverse", NULL);
         xmlSecBnFinalize(&bn);
-        return(-1);
+        return(NULL);
     }
 
+    res = xmlSecBnToDecString(&bn);
+    if (res == NULL) {
+        xmlSecInternalError("xmlSecBnToDecString", NULL);
+        xmlSecBnFinalize(&bn);
+        return(NULL);
+    }
+
+    /* done */
     xmlSecBnFinalize(&bn);
-    return(0);
+    return(res);
 }
 
 static int
-xmlSecMSCngX509IssuerSerialNodeWrite(PCCERT_CONTEXT cert, xmlNodePtr node) {
-    xmlNodePtr cur;
-    xmlNodePtr issuerNameNode;
-    xmlNodePtr issuerNumberNode;
-    xmlChar* buf;
+xmlSecMSCngX509SKIWrite(PCCERT_CONTEXT cert, xmlSecBufferPtr buf) {
+    PCERT_EXTENSION pCertExt;
+    DWORD dwSize;
+    BOOL rv;
     int ret;
 
     xmlSecAssert2(cert != NULL, -1);
-    xmlSecAssert2(node != NULL, -1);
-
-    /* create xml nodes */
-    cur = xmlSecEnsureEmptyChild(node, xmlSecNodeX509IssuerSerial, xmlSecDSigNs);
-    if(cur == NULL) {
-        xmlSecInternalError("xmlSecEnsureEmptyChild", NULL);
-        return(-1);
-    }
-
-    issuerNameNode = xmlSecEnsureEmptyChild(cur, xmlSecNodeX509IssuerName, xmlSecDSigNs);
-    if(issuerNameNode == NULL) {
-        xmlSecInternalError("xmlSecEnsureEmptyChild", NULL);
-        return(-1);
-    }
-
-    issuerNumberNode = xmlSecEnsureEmptyChild(cur, xmlSecNodeX509SerialNumber, xmlSecDSigNs);
-    if(issuerNumberNode == NULL) {
-        xmlSecInternalError("xmlSecEnsureEmptyChild", NULL);
-        return(-1);
-    }
-
-    /* write data */
-    buf = xmlSecMSCngX509NameWrite(&(cert->pCertInfo->Issuer));
-    if(buf == NULL) {
-        xmlSecInternalError("xmlSecMSCngX509NameWrite", NULL);
-        return(-1);
-    }
-
-    ret = xmlSecNodeEncodeAndSetContent(issuerNameNode, buf);
-    if(ret < 0) {
-        xmlSecInternalError("xmlSecNodeEncodeAndSetContent", NULL);
-        xmlFree(buf);
-        return(-1);
-    }
-
-    xmlFree(buf);
-
-    ret = xmlSecMSCngASN1IntegerWrite(issuerNumberNode, &(cert->pCertInfo->SerialNumber));
-    if(ret < 0) {
-        xmlSecInternalError("xmlSecMSCngASN1IntegerWrite", NULL);
-        return(-1);
-    }
-    return(0);
-}
-
-static xmlChar*
-xmlSecMSCngX509SKIWrite(PCCERT_CONTEXT cert) {
-    xmlChar *res = NULL;
-    DWORD dwSize;
-    BYTE *bSKI = NULL;
-    PCERT_EXTENSION pCertExt;
-
-    xmlSecAssert2(cert != NULL, NULL);
+    xmlSecAssert2(buf != NULL, -1);
 
     /* First check if the SKI extension actually exists, otherwise we get a SHA1 hash of the cert */
     pCertExt = CertFindExtension(szOID_SUBJECT_KEY_IDENTIFIER, cert->pCertInfo->cExtension, cert->pCertInfo->rgExtension);
     if (pCertExt == NULL) {
         xmlSecMSCngLastError("CertFindExtension", NULL);
-        return (NULL);
+        return (0);
     }
 
-    if (!CertGetCertificateContextProperty(cert, CERT_KEY_IDENTIFIER_PROP_ID, NULL, &dwSize) || dwSize < 1) {
+    rv = CertGetCertificateContextProperty(cert, CERT_KEY_IDENTIFIER_PROP_ID, NULL, &dwSize);
+    if (!rv || dwSize <= 0) {
         xmlSecMSCngLastError("CertGetCertificateContextProperty", NULL);
-        return (NULL);
-    }
-    bSKI = xmlMalloc(dwSize);
-    if (bSKI == NULL) {
-        xmlSecMallocError(dwSize, NULL);
-        return (NULL);
+        return(-1);
     }
 
-    if (!CertGetCertificateContextProperty(cert, CERT_KEY_IDENTIFIER_PROP_ID, bSKI, &dwSize)) {
+    ret = xmlSecBufferSetMaxSize(buf, dwSize);
+    if (ret < 0) {
+        xmlSecInternalError2("xmlSecBufferSetMaxSize", NULL,
+            "size=%lu", dwSize);
+        return(-1);
+    }
+
+    if (!CertGetCertificateContextProperty(cert, CERT_KEY_IDENTIFIER_PROP_ID, xmlSecBufferGetData(buf), &dwSize)) {
         xmlSecMSCngLastError("CertGetCertificateContextProperty", NULL);
-        xmlFree(bSKI);
-        return (NULL);
-    }
-
-    if (bSKI == NULL) {
-        return(NULL);
-    }
-
-    res = xmlSecBase64Encode(bSKI, dwSize, 0);
-    if(res == NULL) {
-        xmlSecInternalError("xmlSecBase64Encode", NULL);
-        xmlFree(bSKI);
-        return(NULL);
-    }
-    xmlFree(bSKI);
-
-    return(res);
-}
-
-static int
-xmlSecMSCngX509SKINodeWrite(PCCERT_CONTEXT cert, xmlNodePtr node) {
-    xmlChar *buf = NULL;
-    xmlNodePtr cur = NULL;
-    int ret;
-
-    xmlSecAssert2(cert != NULL, -1);
-    xmlSecAssert2(node != NULL, -1);
-
-    buf = xmlSecMSCngX509SKIWrite(cert);
-    if(buf == NULL) {
-        xmlSecInternalError("xmlSecMSCngX509SKIWrite", NULL);
         return(-1);
     }
 
-    cur = xmlSecEnsureEmptyChild(node, xmlSecNodeX509SKI, xmlSecDSigNs);
-    if(cur == NULL) {
-        xmlSecInternalError("xmlSecEnsureEmptyChild", NULL);
-        xmlFree(buf);
+    ret = xmlSecBufferSetSize(buf, dwSize);
+    if (ret < 0) {
+        xmlSecInternalError2("xmlSecBufferSetSize", NULL,
+            "size=%lu", dwSize);
         return(-1);
     }
-
-    ret = xmlSecNodeEncodeAndSetContent(cur, buf);
-    if(ret < 0) {
-        xmlSecInternalError("xmlSecNodeEncodeAndSetContent", NULL);
-        xmlFree(buf);
-        return(-1);
-    }
-
-    xmlFree(buf);
     return(0);
 }
 
+typedef struct _xmlSecMSCngKeyDataX5099WriteContext {
+    HCERTSTORE store;
+    PCCERT_CONTEXT crt;
+    PCCRL_CONTEXT crl;
+    int doneCrts;
+    int doneCrls;
+} xmlSecMSCngKeyDataX5099WriteContext;
+
 static int
-xmlSecMSCngX509CRLNodeWrite(PCCRL_CONTEXT crl, xmlNodePtr node,
-        xmlSecKeyInfoCtxPtr keyInfoCtx) {
-    xmlSecAssert2(crl != NULL, -1);
-    xmlSecAssert2(node != NULL, -1);
-    xmlSecAssert2(keyInfoCtx != NULL, -1);
+xmlSecMSCngKeyDataX509Write(xmlSecKeyDataPtr data, xmlSecKeyValueX509Ptr x509Value,
+                            int content, void* context) {
+    xmlSecMSCngKeyDataX5099WriteContext* ctx;
+    int ret;
 
-    xmlSecNotImplementedError(NULL);
+    xmlSecAssert2(data != NULL, -1);
+    xmlSecAssert2(xmlSecKeyDataCheckId(data, xmlSecMSCngKeyDataX509Id), -1);
+    xmlSecAssert2(x509Value != NULL, -1);
+    xmlSecAssert2(context != NULL, -1);
 
-    return(-1);
+    ctx = (xmlSecMSCngKeyDataX5099WriteContext*)context;
+    xmlSecAssert2(ctx != NULL, -1);
+    xmlSecAssert2(ctx->store != NULL, -1);
+
+    /* try to get and write the next cert if availablle */
+    if (ctx->doneCrts == 0) {
+        ctx->crt = CertEnumCertificatesInStore(ctx->store, ctx->crt);
+        if (ctx->crt != NULL) {
+            if ((content & XMLSEC_X509DATA_CERTIFICATE_NODE) != 0) {
+                xmlSecAssert2(ctx->crt->pbCertEncoded != NULL, -1);
+                xmlSecAssert2(ctx->crt->cbCertEncoded > 0, -1);
+
+                ret = xmlSecBufferSetData(&(x509Value->cert), ctx->crt->pbCertEncoded, ctx->crt->cbCertEncoded);
+                if (ret < 0) {
+                    xmlSecInternalError("xmlSecBufferSetData", xmlSecKeyDataGetName(data));
+                    return(-1);
+                }
+            }
+            if ((content & XMLSEC_X509DATA_SKI_NODE) != 0) {
+                ret = xmlSecMSCngX509SKIWrite(ctx->crt, &(x509Value->ski));
+                if (ret < 0) {
+                    xmlSecInternalError("xmlSecMSCngX509SKIWrite", xmlSecKeyDataGetName(data));
+                    return(-1);
+                }
+            }
+            if ((content & XMLSEC_X509DATA_SUBJECTNAME_NODE) != 0) {
+                xmlSecAssert2(x509Value->subject == NULL, -1);
+                xmlSecAssert2(ctx->crt->pCertInfo != NULL, -1);
+
+                x509Value->subject = xmlSecMSCngX509NameWrite(&(ctx->crt->pCertInfo->Subject));
+                if (x509Value->subject == NULL) {
+                    xmlSecInternalError("xmlSecMSCngX509NameWrite(subject)", xmlSecKeyDataGetName(data));
+                    return(-1);
+                }
+            }
+            if ((content & XMLSEC_X509DATA_ISSUERSERIAL_NODE) != 0) {
+                xmlSecAssert2(x509Value->issuerName == NULL, -1);
+                xmlSecAssert2(x509Value->issuerSerial == NULL, -1);
+                xmlSecAssert2(ctx->crt->pCertInfo != NULL, -1);
+
+                x509Value->issuerName = xmlSecMSCngX509NameWrite(&(ctx->crt->pCertInfo->Issuer));
+                if (x509Value->issuerName == NULL) {
+                    xmlSecInternalError("xmlSecMSCngX509NameWrite(issuer name)", xmlSecKeyDataGetName(data));
+                    return(-1);
+                }
+                x509Value->issuerSerial = xmlSecMSCngASN1IntegerWrite(&(ctx->crt->pCertInfo->SerialNumber));
+                if (x509Value->issuerSerial == NULL) {
+                    xmlSecInternalError("xmlSecMSCngASN1IntegerWrite(issuer serial))", xmlSecKeyDataGetName(data));
+                   return(-1);
+                }
+            }
+            /* done */
+            return(0);
+        } else {
+            ctx->doneCrts = 1;
+        }
+    }
+
+    /* try to get and write the next crl if availablle */
+    if (ctx->doneCrls == 0) {
+        ctx->crl = CertEnumCRLsInStore(ctx->store, ctx->crl);
+        if (ctx->crl != NULL) {
+            if ((content & XMLSEC_X509DATA_CRL_NODE) != 0) {
+                xmlSecAssert2(ctx->crl->pbCrlEncoded != NULL, -1);
+                xmlSecAssert2(ctx->crl->cbCrlEncoded > 0, -1);
+
+                ret = xmlSecBufferSetData(&(x509Value->crl), ctx->crl->pbCrlEncoded, ctx->crl->cbCrlEncoded);
+                if (ret < 0) {
+                    xmlSecInternalError("xmlSecBufferSetData", xmlSecKeyDataGetName(data));
+                    return(-1);
+                }
+            }
+            /* done */
+            return(0);
+        } else {
+            ctx->doneCrls = 1;
+        }
+    }
+
+    /* no more certs or crls */
+    xmlSecAssert2(ctx->doneCrts != 0, -1);
+    xmlSecAssert2(ctx->doneCrls != 0, -1);
+    return(1);
 }
 
 static int
 xmlSecMSCngKeyDataX509XmlWrite(xmlSecKeyDataId id, xmlSecKeyPtr key,
                                xmlNodePtr node, xmlSecKeyInfoCtxPtr keyInfoCtx) {
-    int content;
-    xmlSecKeyDataPtr keyData;
+    xmlSecMSCngKeyDataX5099WriteContext context;
     xmlSecMSCngX509DataCtxPtr x509DataCtx;
-    PCCERT_CONTEXT cert = NULL;
-    HCERTSTORE certs;
-    PCCRL_CONTEXT crlCtx = NULL;
+    xmlSecKeyDataPtr data;
     int ret;
 
     xmlSecAssert2(id == xmlSecMSCngKeyDataX509Id, -1);
     xmlSecAssert2(key != NULL, -1);
-    xmlSecAssert2(node != NULL, -1);
-    xmlSecAssert2(keyInfoCtx != NULL, -1);
 
-    content = xmlSecX509DataGetNodeContent(node, keyInfoCtx);
-    if(content < 0) {
-        xmlSecInternalError("xmlSecX509DataGetNodeContent",
+    /* get x509 data */
+    data = xmlSecKeyGetData(key, id);
+    if (data == NULL) {
+        /* no x509 data in the key */
+        return(0);
+    }
+    x509DataCtx = xmlSecMSCngX509DataGetCtx(data);
+    xmlSecAssert2(x509DataCtx != NULL, -1);
+
+    /* setup context */
+    context.store = x509DataCtx->hMemStore;
+    context.crt = NULL;
+    context.crl = NULL;
+    context.doneCrts = context.doneCrls = 0;
+
+    ret = xmlSecKeyDataX509XmlWrite(data, node, keyInfoCtx,
+        xmlSecBase64GetDefaultLineSize(), 1, /* add line breaks */
+        xmlSecMSCngKeyDataX509Write, &context);
+    if (ret < 0) {
+        xmlSecInternalError("xmlSecKeyDataX509XmlWrite",
             xmlSecKeyDataKlassGetName(id));
         return(-1);
     }
-    if(content == 0) {
-        /* no content -> writer the default */
-        content = XMLSEC_X509DATA_DEFAULT;
-    }
-
-    keyData = xmlSecKeyGetData(key, id);
-    if(keyData == NULL) {
-        /* nothing to do */
-        return(0);
-    }
-
-    xmlSecAssert2(xmlSecKeyDataCheckId(keyData, xmlSecMSCngKeyDataX509Id), -1);
-    x509DataCtx = xmlSecMSCngX509DataGetCtx(keyData);
-    certs = x509DataCtx->hMemStore;
-
-    /* write certificates */
-    while((cert = CertEnumCertificatesInStore(certs, cert)) != NULL) {
-        if((content & XMLSEC_X509DATA_CERTIFICATE_NODE) != 0) {
-            ret = xmlSecMSCngX509CertificateNodeWrite(cert, node, keyInfoCtx);
-            if(ret < 0) {
-                xmlSecInternalError("xmlSecMSCngX509CertificateNodeWrite",
-                    xmlSecKeyDataKlassGetName(id));
-                return(-1);
-            }
-        }
-
-        if((content & XMLSEC_X509DATA_SUBJECTNAME_NODE) != 0) {
-            ret = xmlSecMSCngX509SubjectNameNodeWrite(cert, node);
-            if(ret < 0) {
-                xmlSecInternalError("xmlSecMSCngX509SubjectNameNodeWrite",
-                    xmlSecKeyDataKlassGetName(id));
-                return(-1);
-            }
-        }
-
-        if((content & XMLSEC_X509DATA_ISSUERSERIAL_NODE) != 0) {
-            ret = xmlSecMSCngX509IssuerSerialNodeWrite(cert, node);
-            if(ret< 0) {
-                xmlSecInternalError("xmlSecMSCngX509IssuerSerialNodeWrite",
-                    xmlSecKeyDataKlassGetName(id));
-                return(-1);
-            }
-        }
-
-        if((content & XMLSEC_X509DATA_SKI_NODE) != 0) {
-            ret = xmlSecMSCngX509SKINodeWrite(cert, node);
-            if(ret < 0) {
-                xmlSecInternalError("xmlSecMSCngX509SKINodeWrite",
-                    xmlSecKeyDataKlassGetName(id));
-                return(-1);
-            }
-        }
-    }
-
-    /* write CRLs */
-    while((crlCtx = CertEnumCRLsInStore(certs, crlCtx)) != NULL) {
-        ret = xmlSecMSCngX509CRLNodeWrite(crlCtx, node, keyInfoCtx);
-        if(ret < 0) {
-            xmlSecInternalError("xmlSecMSCngX509CRLNodeWrite",
-                xmlSecKeyDataKlassGetName(id));
-            return(-1);
-        }
-    }
 
+    /* success */
     return(0);
 }
 
